################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

from ctwin32 import (
    ctypes,
    comctl,     # noqa: F401 for initialization of common controls
    user,
    wndcls,
    WM_CREATE,
    WM_SETFOCUS,
    WM_DESTROY,
    WS_BORDER,
    WS_CHILD,
    WS_VISIBLE,
    WS_SYSMENU,
    WS_MINIMIZEBOX,
    )
from ctwin32.wtypes import RECT

MONTHCAL_CLASS = "SysMonthCal32"
MCS_WEEKNUMBERS = 4
MCM_FIRST = 0x1000
MCM_GETMINREQRECT = MCM_FIRST + 9

################################################################################

class CalendarWnd(wndcls.SimpleWnd):

    def on_message(self, msg, wp, lp):

        if msg == WM_CREATE:
            self.cal = wndcls.BaseWnd(
                user.CreateWindowEx(
                    0,
                    MONTHCAL_CLASS,
                    None,
                    WS_BORDER | WS_CHILD | WS_VISIBLE | MCS_WEEKNUMBERS,
                    0, 0, 0, 0, # resize it later
                    self.hwnd,
                    None,
                    self.hinstance(),
                    None
                    )
                )

            # get the size required to show an entire month
            control = RECT()
            self.cal.send_msg(MCM_GETMINREQRECT, 0, ctypes.addressof(control))

            # resize the control and parent
            self.cal.set_pos(None, 0, 0, control.width, control.height, 0)
            self.adjust_window_rect(control)
            return 0

        elif msg == WM_SETFOCUS:
            self.cal.set_focus()
            return 0

        elif msg == WM_DESTROY:
            user.PostQuitMessage(0)
            return 0

        else:
            return self.def_win_proc(msg, wp, lp)

################################################################################

if __name__ == "__main__":

    icon = wndcls.load_ctwin32_ico()
    style = WS_SYSMENU | WS_MINIMIZEBOX
    wnd = CalendarWnd(wndcls.WndCreateParams("Calendar", icon, style))
    wnd.show()

    while msg := user.GetMessage():
        user.TranslateMessage(msg)
        user.DispatchMessage(msg)

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

# This example shows how to receive raw input even if the application is not
# in the foreground.

from ctwin32 import (
    FILE_TYPE_CHAR,
    HID_USAGE_PAGE_GENERIC,
    HID_USAGE_GENERIC_MOUSE,
    HID_USAGE_GENERIC_KEYBOARD,
    HWND_MESSAGE,
    kernel,
    MOD_ALT,
    MOD_CONTROL,
    MOD_SHIFT,
    MOD_WIN,
    RI_KEY_BREAK,
    RI_KEY_E0,
    RI_KEY_E1,
    RI_KEY_TERMSRV_SET_LED,
    RI_KEY_TERMSRV_SHADOW,
    RI_MOUSE_BUTTON_1_DOWN,
    RI_MOUSE_BUTTON_1_UP,
    RI_MOUSE_BUTTON_2_DOWN,
    RI_MOUSE_BUTTON_2_UP,
    RI_MOUSE_BUTTON_3_DOWN,
    RI_MOUSE_BUTTON_3_UP,
    RI_MOUSE_BUTTON_4_DOWN,
    RI_MOUSE_BUTTON_4_UP,
    RI_MOUSE_BUTTON_5_DOWN,
    RI_MOUSE_BUTTON_5_UP,
    RI_MOUSE_WHEEL,
    RIDEV_INPUTSINK,
    RIM_TYPEMOUSE,
    RIM_TYPEHID,
    RIM_TYPEKEYBOARD,
    STD_INPUT_HANDLE,
    user,
    VK_CONTROL,
    VK_LCONTROL,
    VK_RCONTROL,
    VK_END,
    VK_MENU,
    VK_LMENU,
    VK_RMENU,
    VK_SHIFT,
    VK_LSHIFT,
    VK_RSHIFT,
    VK_LWIN,
    VK_RWIN,
    WM_INPUT,
    wndcls,
    )

# press CTRL+SHIFT+ALT+WIN+end to terminate this program
EXIT_MODIFIERS = MOD_CONTROL | MOD_ALT | MOD_SHIFT | MOD_WIN
EXIT_KEY = VK_END

################################################################################

def dump_raw_mouse(d):
    s = d.usButtonFlags
    state = []
    if s & RI_MOUSE_BUTTON_1_DOWN:
        state.append("b1 down")
    if s & RI_MOUSE_BUTTON_1_UP:
        state.append("b1 up")
    if s & RI_MOUSE_BUTTON_2_DOWN:
        state.append("b2 down")
    if s & RI_MOUSE_BUTTON_2_UP:
        state.append("b2 up")
    if s & RI_MOUSE_BUTTON_3_DOWN:
        state.append("b3 down")
    if s & RI_MOUSE_BUTTON_3_UP:
        state.append("b3 up")
    if s & RI_MOUSE_BUTTON_4_DOWN:
        state.append("b4 down")
    if s & RI_MOUSE_BUTTON_4_UP:
        state.append("b4 up")
    if s & RI_MOUSE_BUTTON_5_DOWN:
        state.append("b5 down")
    if s & RI_MOUSE_BUTTON_5_UP:
        state.append("b5 up")
    if s & RI_MOUSE_WHEEL:
        state.append("wheel")
    msg = [
        "MOUSE:",
        f"flags = {d.usFlags:04x},",
        f"state = {','.join(state)},",
        f"data = {d.usButtonData},",
        f"x / y = {d.lLastX} / {d.lLastY}",
        ]
    print(" ".join(msg))

################################################################################

def dump_raw_keyboard(d):
    s = d.Flags
    state = []
    state.append("up" if s & RI_KEY_BREAK else "down")
    if s & RI_KEY_E0:
        state.append("E0")
    if s & RI_KEY_E1:
        state.append("E1")
    if s & RI_KEY_TERMSRV_SET_LED:
        state.append("termserv led")
    if s & RI_KEY_TERMSRV_SHADOW:
        state.append("termserv shadow")
    msg = [
        "KEYBOARD:",
        f"scan = {d.MakeCode:02x},",
        f"state = {','.join(state)},",
        f"vkey = {d.VKey:02x}",
        ]
    print(" ".join(msg))

################################################################################

def dump_raw_hid(d):
    msg = [
        "HID:",
        f"size = {d.dwSizeHid},",
        f"count = {d.dwCount},",
        f"data = {bytes(d.bRawData)}",
        ]
    print(" ".join(msg))

################################################################################

def dump_raw_input(ri):
    if ri.header.dwType == RIM_TYPEMOUSE:
        dump_raw_mouse(ri.data.mouse)
    elif ri.header.dwType == RIM_TYPEKEYBOARD:
        dump_raw_keyboard(ri.data.keyboard)
    elif ri.header.dwType == RIM_TYPEHID:
        dump_raw_hid(ri.data.hid)
    else:
        raise ValueError(f"unknown raw input: {ri.header.dwType}")

################################################################################

class RawInputDest(wndcls.SimpleWnd):
    def __init__(self, wc_params):
        self.modifiers = 0
        self.conin = kernel.GetStdHandle(STD_INPUT_HANDLE)
        if self.conin is not None:
            if kernel.GetFileType(self.conin) != FILE_TYPE_CHAR:
                self.conin = None
        super().__init__(wc_params)

    ############################################################################

    def handle_modifiers(self, key, make):
        if key in (VK_CONTROL, VK_LCONTROL, VK_RCONTROL):
            mod = MOD_CONTROL
        elif key in (VK_MENU, VK_LMENU, VK_RMENU):
            mod = MOD_ALT
        elif key in (VK_SHIFT, VK_LSHIFT, VK_RSHIFT):
            mod = MOD_SHIFT
        elif key in (VK_LWIN, VK_RWIN):
            mod = MOD_WIN
        else:
            return False

        if make:
            self.modifiers |= mod
        else:
            self.modifiers &= ~ mod
        return True

    ############################################################################

    def on_message(self, msg, wp, lp):
        if msg != WM_INPUT:
            return self.def_win_proc(msg, wp, lp)

        # get raw input and dump it
        ri = user.GetRawInputData(lp)
        dump_raw_input(ri)

        if self.conin is not None:
            # remove input from console input buffer
            while num := kernel.GetNumberOfConsoleInputEvents(self.conin):
                kernel.ReadConsoleInput(self.conin, num)

        # exit program if the corresponding input is detected
        if ri.header.dwType == RIM_TYPEKEYBOARD:
            key = ri.data.keyboard.VKey
            make = not (ri.data.keyboard.Flags & RI_KEY_BREAK)
            self.handle_modifiers(key, make)
            if key == EXIT_KEY and self.modifiers == EXIT_MODIFIERS:
                user.PostQuitMessage(0)

        return 0

################################################################################

if __name__ == "__main__":

    wnd = RawInputDest(wndcls.WndCreateParams(parent=HWND_MESSAGE))

    gen = HID_USAGE_PAGE_GENERIC
    mouse = HID_USAGE_GENERIC_MOUSE
    keybo = HID_USAGE_GENERIC_KEYBOARD
    raw_devs = [
        user.RAWINPUTDEVICE(gen, mouse, RIDEV_INPUTSINK, wnd.hwnd),
        user.RAWINPUTDEVICE(gen, keybo, RIDEV_INPUTSINK, wnd.hwnd),
        ]
    user.RegisterRawInputDevices(raw_devs)

    while msg := user.GetMessage():
        user.TranslateMessage(msg)
        user.DispatchMessage(msg)

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import os
import ctypes
import ctwin32
from ctwin32 import (
    suppress_winerr,
    ERROR_MORE_DATA,
    FILE_ATTRIBUTE_REPARSE_POINT,
    FILE_FLAG_OPEN_REPARSE_POINT,
    FILE_FLAG_BACKUP_SEMANTICS,
    IO_REPARSE_TAG_SYMLINK,
    IO_REPARSE_TAG_MOUNT_POINT,
    )
from ctwin32.wtypes import (
    BYTE,
    ULONG,
    USHORT,
    WCHAR,
    )
from ctwin32.kernel import create_file, DeviceIoControl, iter_dir

################################################################################

class RP_DATA_SYMLINK(ctypes.Structure):
    _fields_ = (
        ("SubstituteNameOffset", USHORT),
        ("SubstituteNameLength", USHORT),
        ("PrintNameOffset", USHORT),
        ("PrintNameLength", USHORT),
        ("Flags", ULONG),
        ("PathBuffer", WCHAR * 1),
        )

class RP_DATA_MOUNTPOINT(ctypes.Structure):
    _fields_ = (
        ("SubstituteNameOffset", USHORT),
        ("SubstituteNameLength", USHORT),
        ("PrintNameOffset", USHORT),
        ("PrintNameLength", USHORT),
        ("PathBuffer", WCHAR * 1),
        )

class RP_DATA_GENERIC(ctypes.Structure):
    _fields_ = (
        ("DataBuffer", BYTE * 1),
        )

class RP_DATA_UN(ctypes.Union):
    _fields_ = (
        ("SymbolicLinkReparseBuffer", RP_DATA_SYMLINK),
        ("MountPointReparseBuffer", RP_DATA_MOUNTPOINT),
        ("GenericReparseBuffer", RP_DATA_GENERIC),
        )

class REPARSE_DATA_BUFFER(ctypes.Structure):
    _fields_ = (
        ("ReparseTag", ULONG),
        ("ReparseDataLength", USHORT),
        ("Reserved", USHORT),
        ("_anon", RP_DATA_UN),
        )
    _anonymous_ = ("_anon",)

FSCTL_GET_REPARSE_POINT = 0x000900a8

################################################################################

def readlink(link, get_subst_name=False):
    flags = FILE_FLAG_OPEN_REPARSE_POINT | FILE_FLAG_BACKUP_SEMANTICS
    with create_file(link, 0, 0, flags) as hdl:
        size = 512
        while True:
            with suppress_winerr(ERROR_MORE_DATA):
                buf = DeviceIoControl(hdl, FSCTL_GET_REPARSE_POINT, None, size)
                break
            size *= 2
    rpd = REPARSE_DATA_BUFFER.from_buffer(buf)
    if rpd.ReparseTag == IO_REPARSE_TAG_SYMLINK:
        path_buff = (
            ctypes.addressof(rpd) +
            REPARSE_DATA_BUFFER.SymbolicLinkReparseBuffer.offset +
            RP_DATA_SYMLINK.PathBuffer.offset
            )
        src = rpd.SymbolicLinkReparseBuffer
    elif rpd.ReparseTag == IO_REPARSE_TAG_MOUNT_POINT:
        path_buff = (
            ctypes.addressof(rpd) +
            REPARSE_DATA_BUFFER.MountPointReparseBuffer.offset +
            RP_DATA_MOUNTPOINT.PathBuffer.offset
            )
        src = rpd.MountPointReparseBuffer
    else:
        raise RuntimeError(f"unhandled tag: 0x{rpd.ReparseTag:08x}")
    if get_subst_name:
        saddr = path_buff + src.SubstituteNameOffset
        slen = src.SubstituteNameLength // 2
    else:
        saddr = path_buff + src.PrintNameOffset
        slen = src.PrintNameLength // 2

    return ctypes.wstring_at(saddr, slen)

################################################################################

reparse_tags = {
    v: k for k, v in vars(ctwin32).items() if k.startswith("IO_REPARSE_TAG_")
    }
follow = (IO_REPARSE_TAG_SYMLINK, IO_REPARSE_TAG_MOUNT_POINT)

for directory, info in iter_dir(os.environ["SYSTEMDRIVE"]):
    if info.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT:
        rpp = rf"{directory}\{info.cFileName}"
        if info.dwReserved0 in reparse_tags:
            tag = reparse_tags[info.dwReserved0]
        else:
            tag = f"UNKNOWN: 0x{info.dwReserved0:08x}"
        if info.dwReserved0 in follow:
            print(f"{tag:32} {rpp}\n    -> {readlink(rpp)}")
        else:
            print(f"{tag:32} {rpp}")

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

# This sample prints all the environments, that are available. When run with
# administrative privileges the number of accessible environments will be a
# little higher.

from ctwin32 import (
    kernel,
    ntdll,
    PROCESS_VM_READ,
    PROCESS_QUERY_LIMITED_INFORMATION,
    SE_DEBUG_PRIVILEGE,
    )

try:
    ntdll.RtlAdjustPrivilege(SE_DEBUG_PRIVILEGE, True)
except OSError:
    print("\nFailed to enable debug privilege.")
    print("Several environments will be not be available.\n")

oflg = PROCESS_VM_READ | PROCESS_QUERY_LIMITED_INFORMATION

for np in ntdll.enum_processes():
    print(f"{np.pid:5} {np.name}:")
    try:
        with kernel.OpenProcess(oflg, False, np.pid) as hdl:
            for k,v in sorted(kernel.get_proc_env_as_dict(hdl).items()):
                print(f"    {k} = {v}")
    except OSError as e:
        print(f"    {e}")
    print()



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import ctypes
from .wtypes import (
    BOOL,
    HANDLE,
    PPVOID,
    PVOID,
    )
from . import (
    ref,
    fun_fact,
    raise_on_zero,
    multi_str_from_addr,
    TOKEN_READ,
    kernel,
    advapi,
    )

################################################################################

_ue = ctypes.WinDLL("userenv.dll", use_last_error=True)

_DestroyEnvironmentBlock = fun_fact(
    _ue.DestroyEnvironmentBlock, (BOOL, PVOID)
    )

_CreateEnvironmentBlock = fun_fact(
    _ue.CreateEnvironmentBlock, (BOOL, PPVOID, HANDLE, BOOL)
    )

################################################################################

def _env_block_from_token(token, inherit):
    ptr = PVOID()
    raise_on_zero(_CreateEnvironmentBlock(ref(ptr), token, inherit))
    try:
        return multi_str_from_addr(ptr.value)
    finally:
        raise_on_zero(_DestroyEnvironmentBlock(ptr))

################################################################################

def CreateEnvironmentBlock(token=None, inherit=False):
    if token is None:
        with advapi.OpenProcessToken(
                kernel.GetCurrentProcess(),
                TOKEN_READ
                ) as t:
            return _env_block_from_token(t, inherit)
    else:
        return _env_block_from_token(token, inherit)

################################################################################

def create_env_block_as_dict(token=None, inherit=False):
    return kernel.env_str_to_dict(CreateEnvironmentBlock(token, inherit))

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################
#
# This example demonstrates how to make a console window of a current Windows
# version (Win10 >= 1909?) interpret 'virtual terminal sequences'. It is
# derived from
# https://learn.microsoft.com/en-us/windows/console/console-virtual-terminal-sequences
#
################################################################################

from ctwin32 import kernel

# enable 'virtual terminal sequences'
kernel.enable_virt_term()

print("\x1b[31mThis text has a red foreground using SGR.31.\r\n")
print(
    "\x1b[1mThis text has a bright (bold) red foreground using SGR.1 to "
    "affect the previous color setting.\r\n"
    )
print(
    "\x1b[mThis text has returned to default colors using SGR.0 "
    "implicitly.\r\n"
    )
print(
    "\x1b[34;46mThis text shows the foreground and background change at "
    "the same time.\r\n"
    )
print(
    "\x1b[0mThis text has returned to default colors using SGR.0 "
    "explicitly.\r\n"
    )
print(
    "\x1b[31;32;33;34;35;36;101;102;103;104;105;106;107mThis text attempts "
    "to apply many colors in the same command. Note the colors are applied "
    "from left to right so only the right-most option of foreground cyan "
    "(SGR.36) and background bright white (SGR.107) is effective.\r\n"
    )
print("\x1b[39mThis text has restored the foreground color only.\r\n")
print("\x1b[49mThis text has restored the background color only.\r\n")



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

from types import SimpleNamespace as _namespace

import ctypes
from .wtypes import (
    byte_buffer,
    string_buffer,
    BOOL,
    BYTE,
    CallbackContext,
    CallbackContextPtr,
    DWORD,
    HANDLE,
    HINSTANCE,
    HWND,
    INT,
    INT_PTR,
    LOGFONT,
    LONG,
    LONG_PTR,
    LPARAM,
    LRESULT,
    NTSTATUS,
    PDWORD,
    POINT,
    POINTER,
    PPOINT,
    PPVOID,
    PPWSTR,
    PRECT,
    PUINT,
    PVOID,
    PWSTR,
    SHORT,
    RECT,
    UINT,
    UINT_PTR,
    ULONG,
    WCHAR,
    WinError,
    WPARAM,
    WORD,
    )
from . import (
    ref,
    ntdll,
    kernel,
    raise_if,
    raise_on_zero,
    raise_on_err,
    fun_fact,
    CF_UNICODETEXT,
    GMEM_MOVEABLE,
    GWL_STYLE,
    GWL_EXSTYLE,
    INPUT_KEYBOARD,
    KEYEVENTF_KEYUP,
    LR_DEFAULTSIZE,
    MONITOR_DEFAULTTOPRIMARY,
    RID_INPUT,
    RIM_TYPEHID,
    SPI_GETNONCLIENTMETRICS,
    SPI_SETNONCLIENTMETRICS,
    SPI_GETWHEELSCROLLLINES,
    SPI_SETWHEELSCROLLLINES,
    SPI_GETWORKAREA,
    SPIF_UPDATEINIFILE,
    SPIF_SENDCHANGE,
    SWP_NOSIZE,
    SWP_NOZORDER,
    UOI_FLAGS,
    WAIT_FAILED,
    WM_QUIT,
    WSF_VISIBLE,
    )

_usr = ctypes.WinDLL("user32.dll", use_last_error=True)

################################################################################

_GetWindowThreadProcessId = fun_fact(
    _usr.GetWindowThreadProcessId, (DWORD, HANDLE, PDWORD)
    )

def GetWindowThreadProcessId(hwnd):
    pid = DWORD()
    tid = _GetWindowThreadProcessId(hwnd, ref(pid))
    return tid, pid.value

################################################################################

_IsWindowVisible = fun_fact(_usr.IsWindowVisible, (BOOL, HWND))

def IsWindowVisible(hwnd):
    return bool(_IsWindowVisible(hwnd))

################################################################################

_GetWindowTextLength = fun_fact(
    _usr.GetWindowTextLengthW, (INT, HWND)
    )

def GetWindowTextLength(hwnd):
    return _GetWindowTextLength(hwnd)

################################################################################

_GetWindowText = fun_fact(_usr.GetWindowTextW, (INT, HWND, PWSTR, INT))

def GetWindowText(hwnd):
    slen = GetWindowTextLength(hwnd)
    buf = string_buffer(slen + 1)
    res = _GetWindowText(hwnd, buf, slen + 1)
    raise_if(res != slen)
    return buf.value

################################################################################

_SetWindowText = fun_fact(_usr.SetWindowTextW, (BOOL, HWND, PWSTR))

def SetWindowText(hwnd, txt):
    raise_on_zero(_SetWindowText(hwnd, txt))

################################################################################

_GetClassName = fun_fact(
    _usr.GetClassNameW, (INT, HWND, PWSTR, INT)
    )

def GetClassName(hwnd):
    size = 32
    while True:
        size *= 2
        buf = string_buffer(size)
        res = _GetClassName(hwnd, buf, buf._length_)
        raise_on_zero(res)
        if res != size - 1:
            return buf.value

################################################################################

_GetWindowLong = fun_fact(_usr.GetWindowLongW, (LONG, HWND, INT))

def GetWindowLong(hwnd, idx):
    return _GetWindowLong(hwnd, idx)

################################################################################

_GetWindowLongPtr = fun_fact(
    _usr.GetWindowLongPtrW, (LONG_PTR, HWND, INT)
    )

def GetWindowLongPtr(hwnd, idx):
    return _GetWindowLongPtr(hwnd, idx)

################################################################################

_SetWindowLong = fun_fact(
    _usr.SetWindowLongW, (LONG, HWND, INT, LONG)
    )

def SetWindowLong(hwnd, idx, value):
    return _SetWindowLong(hwnd, idx, value)

################################################################################

_SetWindowLongPtr = fun_fact(
    _usr.SetWindowLongPtrW, (LONG_PTR, HWND, INT, LONG_PTR)
    )

def SetWindowLongPtr(hwnd, idx, value):
    return _SetWindowLongPtr(hwnd, idx, value)

################################################################################

_EnumWindowsCallback = ctypes.WINFUNCTYPE(
    BOOL,
    HWND,
    CallbackContextPtr
    )

@_EnumWindowsCallback
def _EnumWndCb(hwnd, ctxt):
    # cannot propagate exceptions from callback
    with kernel.terminate_on_exception():
        cbc = ctxt.contents
        res = cbc.callback(hwnd, cbc.context)
        # keep on enumerating if the callback fails to return a value
        return res if res is not None else True

################################################################################

_EnumWindows = fun_fact(
    _usr.EnumWindows, (BOOL, _EnumWindowsCallback, CallbackContextPtr)
    )

def EnumWindows(callback, context):
    cbc = CallbackContext(callback, context)
    _EnumWindows(_EnumWndCb, ref(cbc))

################################################################################

_EnumChildWindows = fun_fact(
    _usr.EnumChildWindows,
    (BOOL, HWND, _EnumWindowsCallback, CallbackContextPtr)
    )

def EnumChildWindows(hwnd, callback, context):
    cbc = CallbackContext(callback, context)
    _EnumChildWindows(hwnd, _EnumWndCb, ref(cbc))

################################################################################

_EnumThreadWindows = fun_fact(
    _usr.EnumThreadWindows,
    (BOOL, DWORD, _EnumWindowsCallback, CallbackContextPtr)
    )

def EnumThreadWindows(tid, callback, context):
    cbc = CallbackContext(callback, context)
    _EnumThreadWindows(tid, _EnumWndCb, ref(cbc))

################################################################################

def _get_wnd_lst_cb(hwnd, wnd_lst):
    _, pid = GetWindowThreadProcessId(hwnd)
    d = _namespace(
        hwnd=hwnd,
        text=GetWindowText(hwnd),
        pid=pid,
        pname=ntdll.proc_path_from_pid(pid),
        cls=GetClassName(hwnd),
        style=GetWindowLong(hwnd, GWL_STYLE),
        exstyle=GetWindowLong(hwnd, GWL_EXSTYLE)
        )
    wnd_lst.append(d)
    return True

def get_window_list():
    wnd_lst = []
    EnumWindows(_get_wnd_lst_cb, wnd_lst)
    return wnd_lst

def get_child_window_list(hwnd):
    wnd_lst = []
    EnumChildWindows(hwnd, _get_wnd_lst_cb, wnd_lst)
    return wnd_lst

def get_thread_window_list(tid):
    wnd_lst = []
    EnumThreadWindows(tid, _get_wnd_lst_cb, wnd_lst)
    return wnd_lst

################################################################################

_WaitForInputIdle = fun_fact(
    _usr.WaitForInputIdle, (DWORD, HANDLE, DWORD)
    )

def WaitForInputIdle(proc, timeout):
    res = _WaitForInputIdle(proc, timeout)
    raise_if(res == WAIT_FAILED)
    return res

################################################################################

_PostMessage = fun_fact(
    _usr.PostMessageW,
    (BOOL, HWND, UINT, UINT_PTR, LONG_PTR)
    )

def PostMessage(hwnd, msg, wp, lp):
    raise_on_zero(_PostMessage(hwnd, msg, wp, lp))

################################################################################

PostQuitMessage = fun_fact(_usr.PostQuitMessage, (None, INT))

################################################################################

_SendMessage = fun_fact(
    _usr.SendMessageW,
    (LONG_PTR, HWND, UINT, UINT_PTR, LONG_PTR)
    )

def SendMessage(hwnd, msg, wp, lp):
    return _SendMessage(hwnd, msg, wp, lp)

################################################################################

_SendMessageTimeout = fun_fact(
    _usr.SendMessageTimeoutW, (
        LONG_PTR,
        HWND,
        UINT,
        UINT_PTR,
        LONG_PTR,
        UINT,
        UINT,
        PDWORD
        )
    )

def SendMessageTimeout(hwnd, msg, wp, lp, flags, timeout):
    result = DWORD()
    raise_on_zero(
        _SendMessageTimeout(
            hwnd,
            msg,
            wp,
            lp,
            flags,
            timeout,
            ref(result)
            )
        )
    return result.value

################################################################################

_GetWindow = fun_fact(_usr.GetWindow, (HWND, HWND, UINT))

def GetWindow(hwnd, cmd):
    return _GetWindow(hwnd, cmd)

################################################################################

_GetAsyncKeyState = fun_fact(_usr.GetAsyncKeyState, (SHORT, INT))

def GetAsyncKeyState(vkey):
    return _GetAsyncKeyState(vkey)

################################################################################

_GetWindowRect = fun_fact(_usr.GetWindowRect, (BOOL, HWND, PRECT))

def GetWindowRect(hwnd):
    rc = RECT()
    raise_on_zero(_GetWindowRect(hwnd, ref(rc)))
    return rc

################################################################################

_GetClientRect = fun_fact(_usr.GetClientRect, (BOOL, HWND, PRECT))

def GetClientRect(hwnd):
    rc = RECT()
    raise_on_zero(_GetClientRect(hwnd, ref(rc)))
    return rc

################################################################################

_AdjustWindowRectEx = fun_fact(
    _usr.AdjustWindowRectEx, (BOOL, PRECT, DWORD, BOOL, DWORD)
    )

def AdjustWindowRectEx(rc, style, has_menu, exstyle):
    new_rect = rc.copy()
    raise_on_zero(_AdjustWindowRectEx(ref(new_rect), style, has_menu, exstyle))
    return new_rect

################################################################################

class WINDOWPLACEMENT(ctypes.Structure):
    _fields_ = (
        ("length", UINT),
        ("flags", UINT),
        ("showCmd", UINT),
        ("ptMinPosition", POINT),
        ("ptMaxPosition", POINT),
        ("rcNormalPosition", RECT),
        )

    def __init__(self, f=0, s=1, mi=(0, 0), ma=(0, 0), no=(0, 0, 0, 0)):
        self.length = ctypes.sizeof(WINDOWPLACEMENT)
        self.flags = f
        self.showCmd = s
        self.ptMinPosition = mi
        self.ptMaxPosition = ma
        self.rcNormalPosition = no

    def __repr__(self):
        cl = self.__class__.__name__
        ln = self.length
        fl = self.flags
        sc = self.showCmd
        mi = f"({self.ptMinPosition.x}, {self.ptMinPosition.y})"
        ma = f"({self.ptMaxPosition.x}, {self.ptMaxPosition.y})"
        no = (
            f"({self.rcNormalPosition.left}, {self.rcNormalPosition.top}, ",
            f"{self.rcNormalPosition.right}, {self.rcNormalPosition.bottom})"
            )
        return f"{cl}({ln}, {fl}, {sc}, {mi}, {ma}, {no})"

PWINDOWPLACEMENT = POINTER(WINDOWPLACEMENT)

################################################################################

_GetWindowPlacement = fun_fact(
    _usr.GetWindowPlacement, (BOOL, HWND, PWINDOWPLACEMENT)
    )

def GetWindowPlacement(hwnd):
    wpt = WINDOWPLACEMENT()
    raise_on_zero(_GetWindowPlacement(hwnd, ref(wpt)))
    return wpt

################################################################################

_SetWindowPlacement = fun_fact(
    _usr.SetWindowPlacement, (BOOL, HWND, PWINDOWPLACEMENT)
    )

def SetWindowPlacement(hwnd, wpt):
    raise_on_zero(_SetWindowPlacement(hwnd, ref(wpt)))

################################################################################

_SetWindowPos = fun_fact(
    _usr.SetWindowPos, (BOOL, HWND, HWND, INT, INT, INT, INT, UINT)
    )

def SetWindowPos(hwnd, ins_after, x, y, cx, cy, flags):
    raise_on_zero(_SetWindowPos(hwnd, ins_after, x, y, cx, cy, flags))

################################################################################

_AttachThreadInput = fun_fact(
    _usr.AttachThreadInput, (BOOL, DWORD, DWORD, BOOL)
    )

def AttachThreadInput(id_attach, id_attach_to, do_attach):
    raise_on_zero(_AttachThreadInput(id_attach, id_attach_to, do_attach))

################################################################################

_BringWindowToTop = fun_fact(_usr.BringWindowToTop, (BOOL, HWND))

def BringWindowToTop(hwnd):
    raise_on_zero(_BringWindowToTop(hwnd))

def to_top_maybe_attach(hwnd):
    wnd_id, _ = GetWindowThreadProcessId(hwnd)
    self_id = kernel.GetCurrentThreadId()
    if wnd_id != self_id:
        AttachThreadInput(self_id, wnd_id, True)
    BringWindowToTop(hwnd)
    if wnd_id != self_id:
        AttachThreadInput(self_id, wnd_id, False)

################################################################################

_SetActiveWindow = fun_fact(_usr.SetActiveWindow, (HWND, HWND))

def SetActiveWindow(hwnd):
    return _SetActiveWindow(hwnd)

################################################################################

_MessageBox = fun_fact(
    _usr.MessageBoxW, (INT, HWND, PWSTR, PWSTR, UINT)
    )

def MessageBox(hwnd, text, caption, flags):
    res = _MessageBox(hwnd, text, caption, flags)
    raise_on_zero(res)
    return res

################################################################################

class MOUSEINPUT(ctypes.Structure):
    _fields_ = (
        ("dx", LONG),
        ("dy", LONG),
        ("mouseData", DWORD),
        ("dwFlags", DWORD),
        ("time", DWORD),
        ("dwExtraInfo", UINT_PTR),
        )

class KEYBDINPUT(ctypes.Structure):
    _fields_ = (
        ("wVk", WORD),
        ("wScan", WORD),
        ("dwFlags", DWORD),
        ("time", DWORD),
        ("dwExtraInfo", UINT_PTR),
        )

class HARDWAREINPUT(ctypes.Structure):
    _fields_ = (
        ("uMsg", DWORD),
        ("wParamL", WORD),
        ("wParamH", WORD),
        )

class _DUMMY_INPUT_UNION(ctypes.Union):
    _fields_ = (
        ("mi", MOUSEINPUT),
        ("ki", KEYBDINPUT),
        ("hi", HARDWAREINPUT),
        )

class INPUT(ctypes.Structure):
    _anonymous_ = ("anon",)
    _fields_ = (
        ("type", DWORD),
        ("anon", _DUMMY_INPUT_UNION),
        )

    def copy(self):
        other = INPUT()
        ctypes.memmove(ref(other), ref(self), ctypes.sizeof(INPUT))
        return other

    def as_keyup(self):
        if not self.type == INPUT_KEYBOARD:
            raise ValueError("not INPUT_KEYBOARD")
        up = self.copy()
        up.ki.dwFlags |= KEYEVENTF_KEYUP
        return up

PINPUT = POINTER(INPUT)

################################################################################

def kb_input(vk, scan, flags=0):
    kip = INPUT()
    kip.type = INPUT_KEYBOARD
    kip.ki.wVk = vk
    kip.ki.wScan = scan
    kip.ki.dwFlags = flags
    return kip

################################################################################

_SendInput = fun_fact(_usr.SendInput, (UINT, UINT, PINPUT, INT))

def SendInput(inputs):
    if isinstance(inputs, INPUT):
        num, ptr = 1, ref(inputs)
    else:
        try:
            num = len(inputs)
            if not num:
                return
            inputs = (INPUT * num)(*inputs)
            ptr = ctypes.cast(inputs, PINPUT)
        except Exception as e:
            raise TypeError(f"expected INPUT or list of INPUTs: {e}") from e
    raise_on_zero(_SendInput(num, ptr, ctypes.sizeof(INPUT)))

################################################################################

_ExitWindowsEx = fun_fact(_usr.ExitWindowsEx, (BOOL, UINT, DWORD))

def ExitWindowsEx(flags, reason):
    raise_on_zero(_ExitWindowsEx(flags, reason))

################################################################################

_LockWorkStation = fun_fact(_usr.LockWorkStation, (BOOL,))

def LockWorkStation():
    raise_on_zero(_LockWorkStation())

################################################################################

GetDesktopWindow = fun_fact(_usr.GetDesktopWindow, (HWND,))

################################################################################

GetShellWindow = fun_fact(_usr.GetShellWindow, (HWND,))

################################################################################

_MonitorFromWindow = fun_fact(_usr.MonitorFromWindow, (HANDLE, HWND, DWORD))

def MonitorFromWindow(hwnd, flags=MONITOR_DEFAULTTOPRIMARY):
    return _MonitorFromWindow(hwnd, flags)

################################################################################

class MONITORINFOEX(ctypes.Structure):
    _fields_ = (
        ("cbSize", DWORD),
        ("rcMonitor", RECT),
        ("rcWork", RECT),
        ("dwFlags", DWORD),
        ("szDevice", WCHAR * 32),
        )

    def __init__(self):
        self.cbSize = ctypes.sizeof(self)

PMONITORINFOEX = POINTER(MONITORINFOEX)

################################################################################

_GetMonitorInfo = fun_fact(_usr.GetMonitorInfoW, (BOOL, HANDLE, PMONITORINFOEX))

def GetMonitorInfo(hmon):
    mi = MONITORINFOEX()
    raise_on_zero(_GetMonitorInfo(hmon, ref(mi)))
    return mi

################################################################################

def get_window_center(hwnd=None):
    if hwnd is None:
        return GetMonitorInfo(MonitorFromWindow(None)).rcMonitor.center
    return GetWindowRect(hwnd).center

################################################################################

def center_window(to_be_centered, center_on=None):
    center_x, center_y = get_window_center(center_on)
    rc = GetWindowRect(to_be_centered)
    SetWindowPos(
        to_be_centered,
        None,
        rc.left + center_x - (rc.left + rc.right) // 2,
        rc.top + center_y - (rc.top + rc.bottom) // 2,
        0,
        0,
        SWP_NOSIZE | SWP_NOZORDER
        )

################################################################################

def start_centered(arglist):
    def center_wnd_cb(hwnd, _):
        center_window(hwnd)
        return True

    with kernel.create_process(arglist) as pi:
        WaitForInputIdle(pi.hProcess, 10000)
        EnumThreadWindows(pi.dwThreadId, center_wnd_cb, None)

################################################################################

_LoadString = fun_fact(_usr.LoadStringW, (INT, HANDLE, UINT, PPWSTR, INT))

def LoadString(hinst, strid):
    ptr = PWSTR()
    raise_on_zero(res := _LoadString(hinst, strid, ref(ptr), 0))
    return ctypes.wstring_at(ptr, res)

################################################################################

_LoadCursor = fun_fact(_usr.LoadCursorW, (HANDLE, HANDLE, PWSTR))

def LoadCursor(hinst, cname):
    if isinstance(cname, int) and cname < 2**16:
        cname = ctypes.cast(cname, PWSTR)
    raise_on_zero(res := _LoadCursor(hinst, cname))
    return res

################################################################################

_LoadIcon = fun_fact(_usr.LoadIconW, (HANDLE, HANDLE, PWSTR))

def LoadIcon(hinst, cname):
    if isinstance(cname, int) and cname < 2**16:
        cname = ctypes.cast(cname, PWSTR)
    raise_on_zero(res := _LoadIcon(hinst, cname))
    return res

################################################################################

_DefWindowProc = fun_fact(
    _usr.DefWindowProcW, (LRESULT, HWND, UINT, WPARAM, LPARAM)
    )

def DefWindowProc(hwnd, msg, wp, lp):
    return _DefWindowProc(hwnd, msg, wp, lp)

################################################################################

class CREATESTRUCT(ctypes.Structure):
    _fields_ = (
        ("lpCreateParams", PVOID),
        ("hInstance", HANDLE),
        ("hMenu", HANDLE),
        ("hwndParent", HWND),
        ("cx", INT),
        ("cy", INT),
        ("x", INT),
        ("y", INT),
        ("style", LONG),
        ("lpszName", PWSTR),
        ("lpszClass", PWSTR),
        ("dwExStyle", DWORD),
        )

################################################################################

WNDPROC = ctypes.WINFUNCTYPE(
    LRESULT,
    HWND,
    UINT,
    WPARAM,
    LPARAM
    )

class WNDCLASS(ctypes.Structure):
    _fields_ = (
        ("style", UINT),
        ("lpfnWndProc", WNDPROC),
        ("cbClsExtra", INT),
        ("cbWndExtra", INT),
        ("hInstance", HANDLE),
        ("hIcon", HANDLE),
        ("hCursor", HANDLE),
        ("hbrBackground", HANDLE),
        ("lpszMenuName", PWSTR),
        ("lpszClassName", PWSTR),
        )

PWNDCLASS = POINTER(WNDCLASS)

################################################################################

class MSG(ctypes.Structure):
    _fields_ = (
        ("hWnd", HWND),
        ("message", UINT),
        ("wParam", WPARAM),
        ("lParam", LPARAM),
        ("time", DWORD),
        ("pt", POINT)
        )

    def __bool__(self):
        return self.message != WM_QUIT

PMSG = POINTER(MSG)

################################################################################

class PAINTSTRUCT(ctypes.Structure):
    _fields_ = (
        ("hdc", HANDLE),
        ("fErase", BOOL),
        ("rcPaint", RECT),
        ("fRestore", BOOL),
        ("fIncUpdate", BOOL),
        ("rgbReserved", BYTE * 32),
        )

PPAINTSTRUCT = POINTER(PAINTSTRUCT)

################################################################################

_GetClassInfo = fun_fact(_usr.GetClassInfoW, (BOOL, HANDLE, PWSTR, PWNDCLASS))

def GetClassInfo(hinst, cname):
    wclass = WNDCLASS()
    raise_on_zero(_GetClassInfo(hinst, cname, ref(wclass)))
    return wclass

################################################################################

_RegisterClass = fun_fact(_usr.RegisterClassW, (WORD, PWNDCLASS))

def RegisterClass(wclass):
    res = _RegisterClass(ref(wclass))
    raise_on_zero(res)
    return res

################################################################################

_CreateWindowEx = fun_fact(
    _usr.CreateWindowExW, (
        HWND,
        DWORD,
        PWSTR,
        PWSTR,
        DWORD,
        INT,
        INT,
        INT,
        INT,
        HWND,
        HANDLE,
        HINSTANCE,
        PVOID
        )
    )

def CreateWindowEx(
        ex_style,
        class_name,
        wnd_name,
        style,
        x,
        y,
        width,
        height,
        parent,
        menu,
        hinst,
        create_param
        ):
    hwnd = _CreateWindowEx(
        ex_style,
        class_name,
        wnd_name,
        style,
        x,
        y,
        width,
        height,
        parent,
        menu,
        hinst,
        create_param
        )
    raise_on_zero(hwnd)
    return hwnd

################################################################################

_GetMessage = fun_fact(_usr.GetMessageW, (BOOL, PMSG, HWND, UINT, UINT))

def GetMessage(hwnd=None, msg_min=0, msg_max=0):
    msg = MSG()
    res = _GetMessage(ref(msg), hwnd, msg_min, msg_max)
    raise_if(res == -1)
    return msg

################################################################################

_TranslateMessage = fun_fact(_usr.TranslateMessage, (BOOL, PMSG))

def TranslateMessage(msg):
    return _TranslateMessage(ref(msg))

################################################################################

_DispatchMessage = fun_fact(_usr.DispatchMessageW, (LRESULT, PMSG))

def DispatchMessage(msg):
    return _DispatchMessage(ref(msg))

################################################################################

_ShowWindow = fun_fact(_usr.ShowWindow, (BOOL, HWND, INT))

def ShowWindow(hwnd, cmd):
    return bool(_ShowWindow(hwnd, cmd))

################################################################################

_UpdateWindow = fun_fact(_usr.UpdateWindow, (BOOL, HWND))

def UpdateWindow(hwnd):
    raise_on_zero(_UpdateWindow(hwnd))

################################################################################

_DestroyWindow = fun_fact(_usr.DestroyWindow, (BOOL, HWND))

def DestroyWindow(hwnd):
    raise_on_zero(_DestroyWindow(hwnd))

################################################################################

IsWindow = fun_fact(_usr.IsWindow, (BOOL, HWND))

################################################################################

_GetDlgItem = fun_fact(_usr.GetDlgItem, (HWND, HWND, INT))

def GetDlgItem(hwnd, id):
    res = _GetDlgItem(hwnd, id)
    raise_on_zero(res)
    return res

################################################################################

SendDlgItemMessage = fun_fact(
    _usr.SendDlgItemMessageW, (LRESULT, HWND, INT, UINT, WPARAM, LPARAM)
    )

################################################################################

_SetDlgItemText = fun_fact(
    _usr.SetDlgItemTextW, (BOOL, HWND, INT, PWSTR)
    )

def SetDlgItemText(dlg, id, txt):
    raise_on_zero(_SetDlgItemText(dlg, id, txt))

################################################################################

_GetDlgItemText = fun_fact(
    _usr.GetDlgItemTextW, (UINT, HWND, INT, PWSTR, INT)
    )

def GetDlgItemText(dlg, id):
    length = 128
    res = length
    while res >= length:
        length *= 2
        buf = string_buffer(length)
        kernel.SetLastError(0)
        res = _GetDlgItemText(dlg, id, buf, length)
        raise_on_err(kernel.GetLastError())
    return buf.value

################################################################################

_CheckRadioButton = fun_fact(
    _usr.CheckRadioButton, (BOOL, HWND, INT, INT, INT)
    )

def CheckRadioButton(dlg, first, last, check):
    raise_on_zero(_CheckRadioButton(dlg, first, last, check))

################################################################################

_GetDlgCtrlID = fun_fact(_usr.GetDlgCtrlID, (INT, HWND))

def GetDlgCtrlID(hwnd):
    res = _GetDlgCtrlID(hwnd)
    raise_on_zero(res)
    return res

################################################################################

EnableWindow = fun_fact(_usr.EnableWindow, (BOOL, HWND, BOOL))

################################################################################

SetForegroundWindow = fun_fact(_usr.SetForegroundWindow, (BOOL, HWND))

################################################################################

SetFocus = fun_fact(_usr.SetFocus, (HWND, HWND))

################################################################################

GetParent = fun_fact(_usr.GetParent, (HWND, HWND))

################################################################################

GetMenu = fun_fact(_usr.GetMenu, (HANDLE, HWND))

################################################################################

_InvalidateRect = fun_fact(_usr.InvalidateRect, (BOOL, HWND, PRECT, BOOL))

def InvalidateRect(hwnd, rc, erase):
    prc = ref(rc) if rc is not None else None
    raise_on_zero(_InvalidateRect(hwnd, prc, erase))

################################################################################

WindowFromPoint = fun_fact(_usr.WindowFromPoint, (HWND, POINT))

################################################################################

_MoveWindow = fun_fact(
    _usr.MoveWindow, (
        BOOL,
        HWND,
        INT,
        INT,
        INT,
        INT,
        BOOL
        )
    )

def MoveWindow(hwnd, x, y, width, height, repaint):
    raise_on_zero(_MoveWindow(hwnd, x, y, width, height, repaint))

################################################################################

MapWindowPoints = fun_fact(
    _usr.MapWindowPoints, (
        INT,
        HWND,
        HWND,
        PPOINT,
        UINT,
        )
    )

################################################################################

_GetCursorPos = fun_fact(_usr.GetCursorPos, (BOOL, PPOINT))

def GetCursorPos():
    pt = POINT()
    raise_on_zero(GetCursorPos(ref(pt)))
    return pt

################################################################################

_GetDC = fun_fact(_usr.GetDC, (HANDLE, HWND))

def GetDC(hwnd):
    res = _GetDC(hwnd)
    raise_on_zero(res)
    return res

################################################################################

_GetWindowDC = fun_fact(_usr.GetWindowDC, (HANDLE, HWND))

def GetWindowDC(hwnd):
    res = _GetWindowDC(hwnd)
    raise_on_zero(res)
    return res

################################################################################

_ReleaseDC = fun_fact(_usr.ReleaseDC, (INT, HWND, HANDLE))

def ReleaseDC(hwnd, hdc):
    raise_on_zero(_ReleaseDC(hwnd, hdc))

################################################################################

_SetTimer = fun_fact(_usr.SetTimer, (UINT_PTR, HWND, UINT_PTR, UINT, PVOID))

def SetTimer(hwnd, timer_id, period_ms):
    raise_on_zero(_SetTimer(hwnd, timer_id, period_ms, None))

################################################################################

_KillTimer = fun_fact(_usr.KillTimer, (BOOL, HWND, UINT_PTR))

def KillTimer(hwnd, timer_id):
    raise_on_zero(_KillTimer(hwnd, timer_id))

################################################################################

_CheckDlgButton = fun_fact(_usr.CheckDlgButton, (BOOL, HWND, INT, UINT))

def CheckDlgButton(dlg, id, check):
    raise_on_zero(_CheckDlgButton(dlg, id, check))

################################################################################

IsDlgButtonChecked = fun_fact(_usr.IsDlgButtonChecked, (UINT, HWND, INT))

################################################################################

_BeginPaint = fun_fact(_usr.BeginPaint, (HANDLE, HWND, PPAINTSTRUCT))

def BeginPaint(hwnd):
    ps = PAINTSTRUCT()
    hdc = _BeginPaint(hwnd, ref(ps))
    raise_on_zero(hdc)
    return hdc, ps

################################################################################

_EndPaint = fun_fact(_usr.EndPaint, (BOOL, HWND, PPAINTSTRUCT))

def EndPaint(hwnd, ps):
    raise_on_zero(_EndPaint(hwnd, ref(ps)))

################################################################################

_DrawText = fun_fact(_usr.DrawTextW, (INT, HANDLE, PWSTR, INT, PRECT, UINT))

def DrawText(hdc, txt, rc, fmt):
    raise_on_zero(_DrawText(hdc, txt, len(txt), ref(rc), fmt))

################################################################################

_SetProp = fun_fact(_usr.SetPropW, (BOOL, HWND, PWSTR, HANDLE))

def SetProp(hwnd, name, data):
    raise_on_zero(_SetProp(hwnd, name, data))

################################################################################

_GetProp = fun_fact(_usr.GetPropW, (HANDLE, HWND, PWSTR))

def GetProp(hwnd, name):
    data = _GetProp(hwnd, name)
    raise_on_zero(data)
    return data

def get_prop_def(hwnd, name, default=None):
    data = _GetProp(hwnd, name)
    return data or default

################################################################################

_RemoveProp = fun_fact(_usr.RemovePropW, (HANDLE, HWND, PWSTR))

def RemoveProp(hwnd, name):
    data = _RemoveProp(hwnd, name)
    raise_on_zero(data)
    return data

################################################################################

_EnumPropsCallback = ctypes.WINFUNCTYPE(
    BOOL,
    HWND,
    PVOID,  # cannot use PWSTR, since it can be string or atom
    HANDLE,
    CallbackContextPtr
    )

@_EnumPropsCallback
def _EnumPropsCb(hwnd, name, data, ctxt):
    # cannot propagate exceptions from callback
    with kernel.terminate_on_exception():
        cbc = ctxt.contents
        res = cbc.callback(hwnd, name, data, cbc.context)
        # keep on enumerating if the callback fails to return a value
        return res if res is not None else True

################################################################################

_EnumPropsEx = fun_fact(
    _usr.EnumPropsExW, (INT, HWND, _EnumPropsCallback, CallbackContextPtr)
    )

def EnumPropsEx(hwnd, callback, context):
    cbc = CallbackContext(callback, context)
    _EnumPropsEx(hwnd, _EnumPropsCb, ref(cbc))

################################################################################

def get_prop_dict(hwnd):
    props = {}

    @_EnumPropsCallback
    def collect(hwnd, name, data, not_used):
        #      string            or                      atom
        name = PWSTR(name).value if name >= 0x10000 else f"#{name}"
        props[name] = data
        return True

    _EnumPropsEx(hwnd, collect, None)
    return props

################################################################################

_OpenClipboard = fun_fact(_usr.OpenClipboard, (BOOL, HWND))

def OpenClipboard(hwnd):
    raise_on_zero(_OpenClipboard(hwnd))

################################################################################

_EmptyClipboard = fun_fact(_usr.EmptyClipboard, (BOOL,))

def EmptyClipboard():
    raise_on_zero(_EmptyClipboard())

################################################################################

_SetClipboardData = fun_fact(_usr.SetClipboardData, (HANDLE, UINT, HANDLE))

def SetClipboardData(fmt, hmem):
    res = _SetClipboardData(fmt, hmem)
    raise_on_zero(res)
    return res

################################################################################

_GetClipboardData = fun_fact(_usr.GetClipboardData, (HANDLE, UINT))

def GetClipboardData(fmt):
    res = _GetClipboardData(fmt)
    raise_on_zero(res)
    return res

################################################################################

IsClipboardFormatAvailable = fun_fact(
    _usr.IsClipboardFormatAvailable, (BOOL, UINT)
    )

################################################################################

_CloseClipboard = fun_fact(_usr.CloseClipboard, (BOOL,))

def CloseClipboard():
    raise_on_zero(_CloseClipboard())

################################################################################

_GetClipboardFormatName = fun_fact(
    _usr.GetClipboardFormatNameW, (DWORD, DWORD, PWSTR, DWORD)
    )

def GetClipboardFormatName(fmt_atom):
    bufsize = 1024
    buf = string_buffer(bufsize)
    if _GetClipboardFormatName(fmt_atom, buf, bufsize) == 0:
        raise WinError()
    return buf.value

################################################################################

EnumClipboardFormats = fun_fact(_usr.EnumClipboardFormats, (DWORD, DWORD))

################################################################################

def txt_to_clip(txt, hwnd=None):
    buf = string_buffer(txt)
    size = ctypes.sizeof(buf)
    copied = False
    hcopy = kernel.GlobalAlloc(GMEM_MOVEABLE, size)
    try:
        ctypes.memmove(kernel.GlobalLock(hcopy), buf, size)
        kernel.GlobalUnlock(hcopy)
        OpenClipboard(hwnd)
        try:
            EmptyClipboard()
            SetClipboardData(CF_UNICODETEXT, hcopy)
            copied = True
        finally:
            CloseClipboard()
    finally:
        if not copied:
            kernel.GlobalFree(hcopy)

################################################################################

def txt_from_clip(hwnd=None):
    if not IsClipboardFormatAvailable(CF_UNICODETEXT):
        raise OSError("no clipboard text available")
    OpenClipboard(hwnd)
    try:
        hmem = GetClipboardData(CF_UNICODETEXT)
        txt = ctypes.wstring_at(kernel.GlobalLock(hmem))
        kernel.GlobalUnlock(hmem)
    finally:
        CloseClipboard()
    return txt

################################################################################

GetSystemMetrics = fun_fact(_usr.GetSystemMetrics, (INT, INT))

################################################################################

_ScrollWindow = fun_fact(
    _usr.ScrollWindow, (BOOL, HWND, INT, INT, PRECT, PRECT)
    )

def ScrollWindow(hwnd, x, y, scroll_rect=None, clip_rect=None):
    scroll_rect = ref(scroll_rect) if scroll_rect is not None else None
    clip_rect = ref(clip_rect) if clip_rect is not None else None
    raise_on_zero(_ScrollWindow(hwnd, x, y, scroll_rect, clip_rect))

################################################################################

_GetKeyNameText = fun_fact(_usr.GetKeyNameTextW, (INT, LONG, PWSTR, INT))

def GetKeyNameText(lparam, expect_empty=False):
    size = ret = 32
    while ret >= size - 1:
        size *= 2
        key_name = string_buffer(size)
        ret = _GetKeyNameText(lparam, key_name, size)
        raise_if(not ret and not expect_empty)
    return key_name.value

################################################################################

_CreateIconFromResourceEx = fun_fact(
    _usr.CreateIconFromResourceEx, (
        HANDLE, PVOID, DWORD, BOOL, DWORD, INT, INT, UINT
        )
    )

def CreateIconFromResourceEx(
        data,
        cx=0,
        cy=0,
        is_icon=True,
        default_size=False
        ):
    res = _CreateIconFromResourceEx(
        data,
        len(data),
        is_icon,
        0x00030000,
        cx,
        cy,
        LR_DEFAULTSIZE if default_size else 0
        )
    raise_on_zero(res)
    return res

################################################################################

class GUITHREADINFO(ctypes.Structure):
    _fields_ = (
        ("cbSize",        DWORD),
        ("flags",         DWORD),
        ("hwndActive",    HWND),
        ("hwndFocus",     HWND),
        ("hwndCapture",   HWND),
        ("hwndMenuOwner", HWND),
        ("hwndMoveSize",  HWND),
        ("hwndCaret",     HWND),
        ("rcCaret",       RECT),
        )

    def __init__(self):
        self.cbSize = ctypes.sizeof(self)

PGUITHREADINFO = POINTER(GUITHREADINFO)

_GetGUIThreadInfo = fun_fact(
    _usr.GetGUIThreadInfo, (BOOL, DWORD, PGUITHREADINFO)
    )

def GetGUIThreadInfo(tid=0):
    gti = GUITHREADINFO()
    raise_on_zero(_GetGUIThreadInfo(tid, ref(gti)))
    return gti

################################################################################

_SystemParametersInfo = fun_fact(
    _usr.SystemParametersInfoW, (BOOL, UINT, UINT, PVOID, UINT)
    )

################################################################################

class NONCLIENTMETRICS(ctypes.Structure):
    _fields_ = (
        ("cbSize",             UINT),
        ("iBorderWidth",       INT),
        ("iScrollWidth",       INT),
        ("iScrollHeight",      INT),
        ("iCaptionWidth",      INT),
        ("iCaptionHeight",     INT),
        ("lfCaptionFont",      LOGFONT),
        ("iSmCaptionWidth",    INT),
        ("iSmCaptionHeight",   INT),
        ("lfSmCaptionFont",    LOGFONT),
        ("iMenuWidth",         INT),
        ("iMenuHeight",        INT),
        ("lfMenuFont",         LOGFONT),
        ("lfStatusFont",       LOGFONT),
        ("lfMessageFont",      LOGFONT),
        ("iPaddedBorderWidth", INT),
        )

    def __init__(self):
        self.cbSize = ctypes.sizeof(self)

def get_non_client_metrics():
    ncm = NONCLIENTMETRICS()
    raise_on_zero(
        _SystemParametersInfo(
            SPI_GETNONCLIENTMETRICS,
            ncm.cbSize,
            ref(ncm),
            0
            )
        )
    return ncm

def set_non_client_metrics(ncm, winini=SPIF_UPDATEINIFILE | SPIF_SENDCHANGE):
    ncm.cbSize = ctypes.sizeof(ncm)
    raise_on_zero(
        _SystemParametersInfo(
            SPI_SETNONCLIENTMETRICS,
            ncm.cbSize,
            ref(ncm),
            winini
            )
        )

################################################################################

def get_wheel_scroll_lines():
    lines = UINT()
    raise_on_zero(
        _SystemParametersInfo(
            SPI_GETWHEELSCROLLLINES,
            0,
            ref(lines),
            0
            )
        )
    return lines.value

def set_wheel_scroll_lines(lines, winini=SPIF_UPDATEINIFILE | SPIF_SENDCHANGE):
    raise_on_zero(
        _SystemParametersInfo(
            SPI_SETWHEELSCROLLLINES,
            lines,
            None,
            winini
            )
        )

################################################################################

def get_work_area():
    wa = RECT()
    raise_on_zero(_SystemParametersInfo(SPI_GETWORKAREA, 0, ref(wa), 0))
    return wa

################################################################################

class DLGTEMPLATE(ctypes.Structure):
    _pack_ = 2  # for correct length
    _fields_ = (
        ("style", DWORD),
        ("dwExtendedStyle", DWORD),
        ("cdit", WORD),
        ("x", SHORT),
        ("y", SHORT),
        ("cx", SHORT),
        ("cy", SHORT),
        )

################################################################################

class DLGTEMPLATEEX(ctypes.Structure):
    _pack_ = 2  # for correct length
    _fields_ = (
        ("dlgVer", WORD),
        ("signature", WORD),
        ("helpID", DWORD),
        ("exStyle", DWORD),
        ("style", DWORD),
        ("cDlgItems", WORD),
        ("x", WORD),
        ("y", WORD),
        ("cx", WORD),
        ("cy", WORD),
        )

################################################################################

class DLGITEMTEMPLATE(ctypes.Structure):
    _pack_ = 2  # for correct length
    _fields_ = (
        ("style", DWORD),
        ("exstyle", DWORD),
        ("x", SHORT),
        ("y", SHORT),
        ("cx", SHORT),
        ("cy", SHORT),
        ("id", WORD),
        )

################################################################################

class NMHDR(ctypes.Structure):
    _fields_ = (
        ("hwndFrom", HWND),
        ("idFrom", UINT_PTR),
        ("code", UINT),
        )
PNMHDR = POINTER(NMHDR)

MSDN_FIRST = 0xf060       # ModelesS Dialog
MSDN_LAST = MSDN_FIRST + 50

MSDN_ACTIVATE = MSDN_FIRST + 1

class NM_MSD_ACTIVATE(ctypes.Structure):
    _fields_ = (
        ("hdr", NMHDR),
        ("is_active", BOOL),
        )

MSDN_DESTROY = MSDN_FIRST + 2
NM_MSD_DESTROY = NMHDR

################################################################################

DLGPROC = ctypes.WINFUNCTYPE(
    INT_PTR,
    HWND,
    UINT,
    WPARAM,
    LPARAM
    )

################################################################################

_DialogBoxIndirectParam = fun_fact(
    _usr.DialogBoxIndirectParamW, (INT_PTR, HANDLE, PVOID, HWND, DLGPROC, PVOID)
    )

def DialogBoxIndirectParam(templ, parent, dlg_func, init_param, hinst=None):
    kernel.SetLastError(0)
    res = _DialogBoxIndirectParam(hinst, templ, parent, dlg_func, init_param)
    raise_on_err(kernel.GetLastError())
    return res

################################################################################

_CreateDialogIndirectParam = fun_fact(
    _usr.CreateDialogIndirectParamW, (
        HWND, HANDLE, PVOID, HWND, DLGPROC, PVOID
        )
    )

def CreateDialogIndirectParam(templ, parent, dlg_func, init_param, hinst=None):
    res = _CreateDialogIndirectParam(hinst, templ, parent, dlg_func, init_param)
    raise_on_zero(res)
    return res

################################################################################

_EndDialog = fun_fact(_usr.EndDialog, (BOOL, HWND, INT_PTR))

def EndDialog(hdlg, result):
    raise_on_zero(_EndDialog(hdlg, result))

################################################################################

try:
    _w32 = ctypes.WinDLL("win32u.dll", use_last_error=True)
    _NtUserBuildHwndList = fun_fact(
        _w32.NtUserBuildHwndList, (
            NTSTATUS,
            HANDLE,
            HWND,
            BOOL,
            BOOL,
            DWORD,
            DWORD,
            PPVOID,
            PDWORD
            )
        )

    # build_window_list(0, 0) -> EnumWindows(...)
    # build_window_list(parent, 0) -> EnumChildWindows(parent, ...)
    # build_window_list(0, tid) -> EnumThreadWindows(tid, ...)

    def build_window_list(parent_wnd, thread_id, hdesk=0, hide_immersive=True):
        enum_children = bool(parent_wnd)
        allocated = 512
        while True:
            array = (allocated * PVOID)()
            received = DWORD()
            status = _NtUserBuildHwndList(
                hdesk,
                parent_wnd,
                enum_children,
                hide_immersive,
                thread_id,
                allocated,
                array,
                ref(received)
                )
            received = received.value
            if status == 0:
                break

            if status == ntdll.STATUS_BUFFER_TOO_SMALL:
                # avoid under-allocating due to newly added windows -> + 32
                allocated = received + 32
            else:
                raise WinError(ntdll.RtlNtStatusToDosError(status))

        return array[:received - 1]

except (FileNotFoundError, AttributeError):
    def build_window_list(parent_wnd, thread_id, hdesk=0, hide_immersive=True):
        raise NotImplementedError

################################################################################

class _WINDOWCOMPOSITIONATTRIBDATA(ctypes.Structure):
    _fields_ = (
        ("Attrib", UINT),
        ("pvData", PVOID),
        ("cbData", UINT),
        )

def is_window_cloaked(hwnd):
    res = BOOL(0)
    wcad = _WINDOWCOMPOSITIONATTRIBDATA(
        0x12, # counterpart for DWMWA_CLOAKED, no need for a detour via DWM
        ctypes.addressof(res),
        ctypes.sizeof(res)
        )
    try: # noqa: SIM105
        raise_on_zero(_usr.GetWindowCompositionAttribute(hwnd, ref(wcad)))
    except AttributeError:
        pass
    return res.value != 0

################################################################################

_GetProcessWindowStation = fun_fact(_usr.GetProcessWindowStation, (HANDLE,))

def GetProcessWindowStation():
    res = _GetProcessWindowStation()
    raise_on_zero(res)
    return res

################################################################################

_GetThreadDesktop = fun_fact(_usr.GetThreadDesktop, (HANDLE, DWORD))

def GetThreadDesktop(tid):
    res = _GetThreadDesktop(tid)
    raise_on_zero(res)
    return res

################################################################################

_GetUserObjectInformation = fun_fact(
    _usr.GetUserObjectInformationW,
    (BOOL, HANDLE, INT, PVOID, DWORD, PDWORD)
    )

def GetUserObjectInformation(hdl, idx):
    size = DWORD()
    _GetUserObjectInformation(hdl, idx, None, 0, ref(size))
    buf = byte_buffer(size.value)
    raise_on_zero(_GetUserObjectInformation(hdl, idx, buf, size, ref(size)))
    return buf

################################################################################

class USEROBJECTFLAGS(ctypes.Structure):
    _fields_ = (
        ("fInherit", BOOL),
        ("fReserved", BOOL),
        ("dwFlags", DWORD),
        )

def is_interactive_process():
    uof = USEROBJECTFLAGS.from_buffer(
        GetUserObjectInformation(GetProcessWindowStation(), UOI_FLAGS)
        )
    return bool(uof.dwFlags & WSF_VISIBLE)

################################################################################

class RAWINPUTHEADER(ctypes.Structure):
    _fields_ = (
        ("dwType", DWORD),
        ("dwSize", DWORD),
        ("hDevice", HANDLE),
        ("wParam", WPARAM),
        )

class ButtonFlagsData(ctypes.Structure):
    _fields_ = (
        ("usButtonFlags", WORD),
        ("usButtonData", SHORT),
        )

class ButtonFlagsDataUnion(ctypes.Union):
    _anonymous_ = ("anon",)
    _fields_ = (
        ("ulButtons", ULONG),
        ("anon", ButtonFlagsData),
        )

class RAWMOUSE(ctypes.Structure):
    _anonymous_ = ("anon",)
    _fields_ = (
        ("usFlags", WORD),
        ("anon", ButtonFlagsDataUnion),
        ("ulRawButtons", ULONG),
        ("lLastX", LONG),
        ("lLastY", LONG),
        ("ulExtraInformation", ULONG),
        )
class RAWKEYBOARD(ctypes.Structure):
    _fields_ = (
        ("MakeCode", WORD),
        ("Flags", WORD),
        ("Reserved", WORD),
        ("VKey", WORD),
        ("Message", UINT),
        ("ExtraInformation", ULONG),
        )

class RAWHID(ctypes.Structure):
    _fields_ = (
        ("dwSizeHid", DWORD),
        ("dwCount", DWORD),
        ("bRawData", BYTE * 1),
        )

class RAWINPUTDATA(ctypes.Union):
    _fields_ = (
        ("mouse", RAWMOUSE),
        ("keyboard", RAWKEYBOARD),
        ("hid", RAWHID),
        )

class RAWINPUT(ctypes.Structure):
    _fields_ = (
        ("header", RAWINPUTHEADER),
        ("data", RAWINPUTDATA),
        )

class RAWINPUTDEVICE(ctypes.Structure):
    _fields_ = (
        ("usUsagePage", WORD),
        ("usUsage", WORD),
        ("dwFlags", DWORD),
        ("hwndTarget", HWND),
        )

################################################################################

_RegisterRawInputDevices = fun_fact(
    _usr.RegisterRawInputDevices, (BOOL, PVOID, UINT, UINT)
    )

def RegisterRawInputDevices(raw_imp_devs):
    size = len(raw_imp_devs)
    arr = (RAWINPUTDEVICE * size)(*raw_imp_devs)
    raise_on_zero(
        _RegisterRawInputDevices(ref(arr), size, ctypes.sizeof(RAWINPUTDEVICE))
        )

################################################################################

def _make_raw_hid_input(buf, size, count):
    class REAL_RAWHID(ctypes.Structure):
        _fields_ = (
            ("dwSizeHid", DWORD),
            ("dwCount", DWORD),
            ("bRawData", BYTE * (size * count)),
            )
    class REAL_RAWINPUTDATA(ctypes.Union):
        _fields_ = (
            ("mouse", RAWMOUSE),
            ("keyboard", RAWKEYBOARD),
            ("hid", REAL_RAWHID),
            )

    class REAL_RAWINPUT(ctypes.Structure):
        _fields_ = (
            ("header", RAWINPUTHEADER),
            ("data", REAL_RAWINPUTDATA),
            )

    return REAL_RAWINPUT.from_buffer_copy(buf)

################################################################################

_GetRawInputData = fun_fact(
    _usr.GetRawInputData, (UINT, HANDLE, UINT, PVOID, PUINT, UINT)
    )

def GetRawInputData(hRaw):
    size = UINT(0)
    res = _GetRawInputData(
        hRaw,
        RID_INPUT,
        None,
        ref(size),
        ctypes.sizeof(RAWINPUTHEADER)
        )
    raise_if(res != 0)

    # use at least enough as needed by RAWINPUT (keyboard data will be shorter)
    size.value = max(size.value, ctypes.sizeof(RAWINPUT))
    buf = byte_buffer(size.value)
    res = _GetRawInputData(
        hRaw,
        RID_INPUT,
        ref(buf),
        ref(size),
        ctypes.sizeof(RAWINPUTHEADER)
        )
    raise_if(res < 1)

    ri = RAWINPUT.from_buffer_copy(buf)
    if ri.header.dwType == RIM_TYPEHID:
        size, count = ri.data.hid.dwSizeHid, ri.data.hid.dwCount
        ri = _make_raw_hid_input(buf, size, count)
    return ri

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

# This sample demonstrates how to get the version information from the resource
# resource section of a PE file. It simply takes the python executuble as its
# test object.

import sys
import collections
import ctwin32.version

################################################################################

def extract_py_ver(fver_ms, fver_ls):
    major = fver_ms >> 16
    minor = fver_ms & 0xffff
    field3 = fver_ls >> 16
    api_ver = fver_ls & 0xffff

    # see https://github.com/python/cpython/blob/main/PCbuild/field3.py
    micro, remainder = divmod(field3, 1000)
    level, serial = divmod(remainder, 10)
    level = {
        0xa: "alpha",
        0xb: "beta",
        0xc: "candidate",
        0xf: "final"
        }[level]
    ver_info = collections.namedtuple(
        "version_info",
        ["major", "minor", "micro", "releaselevel", "serial"]
        )(major, minor, micro, level, serial)
    return ver_info, api_ver

################################################################################

if __name__ == "__main__":

    # get binary version info from the executable file
    bin_info = ctwin32.version.get_binary_info(sys.executable)

    # extract the interesting values
    extr_ver_info, extr_api_ver = extract_py_ver(
        bin_info.dwFileVersionMS,
        bin_info.dwFileVersionLS
        )
    print(extr_ver_info)

    # compare those values with infos from sys
    if extr_ver_info != sys.version_info:
        raise ValueError("extracted version does not match sys.version_info")
    if extr_api_ver != sys.api_version:
        raise ValueError("extracted API version does not match sys.api_version")

    # just print the string infos
    print()
    str_info = ctwin32.version.get_string_info(sys.executable)
    for i in sorted(str_info):
        print(f"{i:16} : {str_info[i]}")

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################
#
# This script lists the DLL dependencies of an executable PE file.
#
################################################################################

import os
from pathlib import Path
from argparse import ArgumentParser
from types import SimpleNamespace as nspace

from ctwin32 import (
    DIRECTORY_QUERY,
    IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT,
    IMAGE_DIRECTORY_ENTRY_IMPORT,
    kernel,
    ntdll,
    )
from ctwin32.pemap import (
    ApiSet,
    dlattrRva,
    IMAGE_IMPORT_DESCRIPTOR,
    IMAGE_DELAYLOAD_DESCRIPTOR,
    pemap,
    )

API_SET = ApiSet()

################################################################################

def get_known_dlls(for_wow=False):
    sdir = Path(
        kernel.GetSystemDirectory() if not for_wow
        else kernel.GetSystemWow64Directory()
        )
    oa = ntdll.obj_attr(r"\KnownDlls32" if for_wow else r"\KnownDlls")
    with ntdll.NtOpenDirectoryObject(DIRECTORY_QUERY, oa) as hdir:
        name_type_lst = ntdll.NtQueryDirectoryObject(hdir)
    return {
        name.lower(): str(sdir / name)
        for name, typ in sorted(name_type_lst) if typ == "Section"
        }

################################################################################

def get_dll_search_dirs(mod_path, for_wow=False):
    nodup = set()
    result = []
    def append(sd):
        sd = str(sd)
        sdl = sd.lower()
        if sdl not in nodup:
            nodup.add(sdl)
            result.append(sd)
    append(mod_path)
    append(kernel.GetCurrentDirectory())
    append(
        kernel.GetSystemDirectory() if not for_wow
        else kernel.GetSystemWow64Directory()
        )
    append(kernel.GetWindowsDirectory())
    for p in os.environ["PATH"].split(";"):
        append(p)

    return result

################################################################################

class Imports():
    def __init__(self, key, path):
        self.tree = nspace(name=key, delayed=False, children=[])
        self.mods = {key: nspace(elem=self.tree, path=path, resolved=False)}

    def add_dll(self, dependant, dll, path, is_delayed):
        dep = self.mods[dependant.lower()].elem
        dll = dll.lower()
        if dll in (c.name for c in dep.children):
            return
        child = nspace(name=dll, delayed=is_delayed, children=[])
        dep.children.append(child)
        if dll not in self.mods:
            if path is None:
                path = "<NOT FOUND>"
                resolved = True
            else:
                resolved = False
            self.mods[dll] = nspace(elem=child, path=path, resolved=resolved)

    def get_unresolved(self):
        return [m.path for m in self.mods.values() if not m.resolved]

    def mark_resolved(self, key):
        self.mods[key].resolved = True

    def _fmt_nodes(self, node=None, level=0, indent=4):
        res = []
        if node is None:
            node = self.tree
        d = " (delayed)" if node.delayed else ""
        res.append(f"{' ' * level * indent}{node.name}{d}")
        for c in node.children:
            res.extend(self._fmt_nodes(c, level + 1, indent))
        return res

    def get_results(self):
        lines = self._fmt_nodes()
        mlst = [(k, v.path) for k, v in self.mods.items()]
        return mlst, self.tree, "\n".join(lines)

################################################################################

def find_dll_path(name, search_info):
    known_dlls, search_dirs = search_info
    try:
        return known_dlls[name.lower()]
    except KeyError:
        for d in search_dirs:
            p = Path(d) / name
            if p.exists():
                return str(p)
        return None

################################################################################

def get_imported_mod(pe, rva_name, search_info):
    mod_name = pe.bstring(pe.offs_from_rva(rva_name)).lower()
    if api := API_SET.lookup(mod_name):
        mod_name = api
    mpath = find_dll_path(mod_name, search_info)
    if mpath is None and "." not in mod_name:
        mname = mod_name + ".dll"
        if mpath := find_dll_path(mname, search_info):
            mod_name = mname
    return mod_name, mpath

################################################################################

def add_static_imports(imports, pe, search_info):
    im_rva, _ = pe.img_dir(IMAGE_DIRECTORY_ENTRY_IMPORT)
    if not im_rva:
        return
    iid_offs = pe.offs_from_rva(im_rva)
    idx = 0
    while True:
        iid = pe.ctypes_obj(IMAGE_IMPORT_DESCRIPTOR, iid_offs, idx)
        idx += 1
        if iid.TimeDateStamp == 0 and iid.Name == 0:
            # end of imports
            break
        dll, path = get_imported_mod(pe, iid.Name, search_info)
        imports.add_dll(pe.key, dll, path, False)

################################################################################

def add_delay_imports(imports, pe, search_info):
    rva, _ = pe.img_dir(IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT)
    if not rva:
        return
    offs = pe.offs_from_rva(rva)
    if not offs:
        return
    idx = 0
    while True:
        idd = pe.ctypes_obj(IMAGE_DELAYLOAD_DESCRIPTOR, offs, idx)
        idx += 1
        if idd.DllNameRVA == 0 or (idd.Attributes & dlattrRva) == 0:
            break
        dll, path = get_imported_mod(pe, idd.DllNameRVA, search_info)
        imports.add_dll(pe.key, dll, path, True)

################################################################################

def get_dep_tree(filename, delayed, static):
    level = 0
    with pemap(filename) as pe:
        imports = Imports(pe.key, pe.name)
        search_info = (
            get_known_dlls(pe.is_wow()),
            get_dll_search_dirs(Path(pe.name).parent, pe.is_wow())
            )
    while unresolved := imports.get_unresolved():
        for path in unresolved:
            with pemap(path) as pe:
                add_static_imports(imports, pe, search_info)
                if not static and (delayed or level < 1):
                    add_delay_imports(imports, pe, search_info)
                imports.mark_resolved(pe.key)
        level += 1
    return imports.get_results()

################################################################################

def dep_tree():
    ape = ArgumentParser(description="determine dependency tree")
    group = ape.add_mutually_exclusive_group()
    g = group.add_argument
    dhlp = "include delayed imports beyond first level"
    g("-d", "--delay", action="store_true", help=dhlp)
    g("-s", "--static", action="store_true", help="only report static imports")
    ape.add_argument("filename", help="name of file to examine")
    a = ape.parse_args()

    mod_list, _, formatted_tree = get_dep_tree(a.filename, a.delay, a.static)

    max_name_len = max(len(n) for n, _ in mod_list) + 1
    print(f"\nlist of modules ({len(mod_list)}):\n")
    for name, path in mod_list:
        print(f"{name:{max_name_len}}: {path}")
    print("\n\ndependency tree:\n")
    print(formatted_tree)

################################################################################

if __name__ == "__main__":
    dep_tree()

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import sys
import lzma
import base64

import ctypes
from .wtypes import (
    string_buffer,
    HANDLE,
    HWND,
    LPARAM,
    POINT,
    PPOINT,
    PVOID,
    UINT,
    WORD,
    )
from . import (
    ref,
    kernel,
    user,
    gdi,
    BST_CHECKED,
    BST_UNCHECKED,
    BS_DEFPUSHBUTTON,
    COLOR_WINDOW,
    CS_DBLCLKS,
    CS_HREDRAW,
    CS_VREDRAW,
    CW_USEDEFAULT,
    DWLP_MSGRESULT,
    DS_MODALFRAME,
    DS_SETFONT,
    EM_SETPASSWORDCHAR,
    ES_AUTOHSCROLL,
    ES_LEFT,
    GWLP_HINSTANCE,
    GWL_EXSTYLE,
    GWL_STYLE,
    HIWORD,
    HORZRES,
    HORZSIZE,
    HWND_TOP,
    HWND_TOPMOST,
    IDCANCEL,
    IDC_ARROW,
    IDOK,
    LOWORD,
    SS_CENTER,
    SWP_NOACTIVATE,
    SWP_NOMOVE,
    SWP_NOSIZE,
    SWP_NOZORDER,
    SW_SHOW,
    SW_HIDE,
    WA_INACTIVE,
    WM_ACTIVATE,
    WM_COMMAND,
    WM_GETFONT,
    WM_INITDIALOG,
    WM_NCCREATE,
    WM_NCDESTROY,
    WM_NOTIFY,
    WS_BORDER,
    WS_CAPTION,
    WS_CHILD,
    WS_GROUP,
    WS_OVERLAPPEDWINDOW,
    WS_POPUP,
    WS_SYSMENU,
    WS_TABSTOP,
    WS_VISIBLE,
    )

################################################################################

class BaseWnd:

    def __init__(self, hwnd=None):
        self.hwnd = hwnd

    def def_win_proc(self, msg, wp, lp):
        return user.DefWindowProc(self.hwnd, msg, wp, lp)

    def is_window(self):
        return bool(user.IsWindow(self.hwnd))

    def get_dlg_item(self, id):
        return BaseWnd(user.GetDlgItem(self.hwnd, id))

    def send_msg(self, msg, wp, lp):
        return user.SendMessage(self.hwnd, msg, wp, lp)

    def post_msg(self, msg, wp, lp):
        user.PostMessage(self.hwnd, msg, wp, lp)

    def send_dlg_item_msg(self, id, msg, wp, lp):
        return user.SendDlgItemMessage(self.hwnd, id, msg, wp, lp)

    def set_dlg_item_text(self, id, txt):
        user.SetDlgItemText(self.hwnd, id, txt)

    def get_dlg_item_text(self, id):
        return user.GetDlgItemText(self.hwnd, id)

    def send_notify(self, nmhdr):
        return user.SendMessage(
            self.hwnd,
            WM_NOTIFY,
            nmhdr.idFrom,
            LPARAM(ctypes.cast(ref(nmhdr), PVOID).value)
            )

    def destroy(self):
        if self.is_window():
            user.DestroyWindow(self.hwnd)
            self.hwnd = None

    def show(self, how=SW_SHOW):
        user.ShowWindow(self.hwnd, how)

    def hide(self):
        user.ShowWindow(self.hwnd, SW_HIDE)

    def enable(self, enabled=True):
        user.EnableWindow(self.hwnd, enabled)

    def disable(self):
        user.EnableWindow(self.hwnd, False)

    def activate(self):
        user.SetActiveWindow(self.hwnd)

    def set_foreground(self):
        user.SetForegroundWindow(self.hwnd)

    def set_focus(self):
        return BaseWnd(user.SetFocus(self.hwnd))

    def invalidate_rect(self, rc=None, erase=False):
        user.InvalidateRect(self.hwnd, rc, erase)

    def update(self):
        user.UpdateWindow(self.hwnd)

    def get_parent(self):
        return BaseWnd(user.GetParent(self.hwnd))

    def get_menu(self):
        return user.GetMenu(self.hwnd)

    def move(self, rc, repaint=True):
        user.MoveWindow(
            self.hwnd,
            rc.left,
            rc.top,
            rc.width(),
            rc.height(),
            repaint
            )

    def set_pos(self, wnd_ins_after, x, y, cx, cy, flags):
        user.SetWindowPos(
            self.hwnd,
            None if wnd_ins_after is None else wnd_ins_after.hwnd,
            x,
            y,
            cx,
            cy,
            flags
            )

    def center(self, center_on=None):
        user.center_window(self.hwnd, center_on.hwnd if center_on else None)

    def set_topmost(self):
        self.set_pos(
            BaseWnd(HWND_TOPMOST),
            0,
            0,
            0,
            0,
            SWP_NOSIZE | SWP_NOMOVE | SWP_NOACTIVATE
            )

    def set_non_topmost(self):
        self.set_pos(
            BaseWnd(HWND_TOP),
            0,
            0,
            0,
            0,
            SWP_NOSIZE | SWP_NOMOVE | SWP_NOACTIVATE
            )

    def bring_to_top(self):
        user.BringWindowToTop(self.hwnd)

    def map_window_point(self, to_bwnd, pt):
        user.MapWindowPoints(self.hwnd, to_bwnd.hwnd, ref(pt), 1)
        return pt

    def map_window_rect(self, to_bwnd, rc):
        ppt = ctypes.cast(ref(rc), PPOINT)
        user.MapWindowPoints(self.hwnd, to_bwnd.hwnd, ppt, 2)
        return rc

    def lp_pt_to_parent(self, lp):
        pt = POINT.from_lparam(lp)
        user.MapWindowPoints(self.hwnd, user.GetParent(self.hwnd), ref(pt), 1)
        return pt.as_lparam()

    def client_to_screen(self, pt_or_rc):
        meth = (
            self.map_window_point if isinstance(pt_or_rc, POINT)
            else self.map_window_rect
            )
        return meth(BaseWnd(), pt_or_rc)

    def window_rect(self):
        return user.GetWindowRect(self.hwnd)

    def client_rect(self):
        return user.GetClientRect(self.hwnd)

    def adjust_window_rect(self, rc):
        frame = user.AdjustWindowRectEx(
            rc,
            self.get_style(),
            bool(self.get_menu()),
            self.get_exstyle()
            )
        self.set_pos(None, 0, 0, frame.width, frame.height, SWP_NOMOVE)

    def window_rect_as_client(self):
        rc = self.window_rect()
        ppt = ctypes.cast(ref(rc), PPOINT)
        user.MapWindowPoints(None, self.hwnd, ppt, 2)
        return rc

    def window_rect_as_other_client(self, other):
        rc = self.window_rect()
        ppt = ctypes.cast(ref(rc), PPOINT)
        user.MapWindowPoints(None, other.hwnd, ppt, 2)
        return rc

    def get_cursor_pos(self):
        pt = user.GetCursorPos()
        user.MapWindowPoints(None, self.hwnd, ref(pt), 1)
        return pt

    def get_nc_cursor_pos(self):
        pt = user.GetCursorPos()
        rc = self.window_rect()
        pt.x -= rc.left
        pt.y -= rc.top
        return pt

    def cursor_over_thread_wnd(self):
        pt = user.GetCursorPos()
        tid, _ = user.GetWindowThreadProcessId(user.WindowFromPoint(pt))
        return (tid == kernel.GetCurrentThreadId())

    def get_dc(self):
        return user.GetDC(self.hwnd)

    def release_dc(self, hdc):
        user.ReleaseDC(self.hwnd, hdc)

    def get_dpi_scale_100(self):
        hdc = self.get_dc()
        try:
            hres = gdi.GetDeviceCaps(hdc, HORZRES)
            hsize = gdi.GetDeviceCaps(hdc, HORZSIZE)
            dpi_100 = (hres * 2540 + hsize // 2) // hsize
            return (dpi_100 + 48) // 96
        finally:
            self.release_dc(hdc)

    def get_style(self):
        return user.GetWindowLong(self.hwnd, GWL_STYLE)

    def get_exstyle(self):
        return user.GetWindowLong(self.hwnd, GWL_EXSTYLE)

    def modify_style(self, remove, add, flags=0, idx=GWL_STYLE):
        style = user.GetWindowLong(self.hwnd, idx)
        new_style = (style & ~remove) | add
        if new_style != style:
            user.SetWindowLong(self.hwnd, idx, new_style)
            if flags:
                flags |= SWP_NOSIZE | SWP_NOMOVE | SWP_NOZORDER
                self.set_pos(None, 0, 0, 0, 0, flags)

    def modify_exstyle(self, remove, add, flags=0):
        self.modify_style(remove, add, flags, GWL_EXSTYLE)

    def hinstance(self):
        return HANDLE(user.GetWindowLongPtr(self.hwnd, GWLP_HINSTANCE))

    def set_timer(self, id, period):
        user.SetTimer(self.hwnd, id, period)

    def kill_timer(self, id):
        user.KillTimer(self.hwnd, id)

    def set_text(self, txt):
        user.SetWindowText(self.hwnd, txt)

    def get_text(self):
        return user.GetWindowText(self.hwnd)

    def get_font(self):
        return HANDLE(self.send_msg(WM_GETFONT, 0, 0))

    def check_dlg_button(self, id, checked):
        user.CheckDlgButton(
            self.hwnd,
            id,
            BST_CHECKED if checked else BST_UNCHECKED
            )

    def is_dlg_button_checked(self, id):
        return (user.IsDlgButtonChecked(self.hwnd, id) == BST_CHECKED)

    def check_radio_button(self, first, last, check):
        user.CheckRadioButton(self.hwnd, first, last, check)

    def begin_paint(self):
        return user.BeginPaint(self.hwnd)

    def end_paint(self, ps):
        user.EndPaint(self.hwnd, ps)

    def set_prop(self, name, data):
        user.SetProp(self.hwnd, name, data)

    def get_prop(self, name):
        return user.GetProp(self.hwnd, name)

    def get_prop_def(self, name, default=None):
        return user.get_prop_def(self.hwnd, name, default)

    def del_prop(self, name):
        return user.RemoveProp(self.hwnd, name)

################################################################################
################################################################################
################################################################################

class WndCreateParams:
    def __init__(self, name="", icon=0, style=WS_OVERLAPPEDWINDOW, parent=0):
        self.cls = user.WNDCLASS(
            style=CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS,
            hInstance=kernel.GetModuleHandle(None),
            hIcon=icon,
            hCursor=user.LoadCursor(None, IDC_ARROW),
            hbrBackground=COLOR_WINDOW + 1,
            )
        self.wnd_style = style
        self.ex_style = 0
        self.left = self.top = self.width = self.height = CW_USEDEFAULT
        self.menu = None
        self.parent = parent
        self.name = name

################################################################################

_PROP_SELF = kernel.global_add_atom("ctwin32:SimpleWnd:self")

class SimpleWnd(BaseWnd):

    def __init__(self, wc_params: WndCreateParams = None):
        self.hwnd = None
        self.parent = None
        if wc_params is not None:
            self.create(wc_params)

    ############################################################################

    def parent_hwnd(self):
        return self.parent.hwnd if self.parent else None

    ############################################################################

    @user.WNDPROC
    @staticmethod
    def _wnd_proc_(hwnd, msg, wp, lp):
        # Since this is a python callback that ctypes calls when requested
        # by foreign C code, ctypes has no way of propagating any exception
        # back to the python interpreter - those would simply be ignored.
        # Therefore any exception has to terminate the process.
        with kernel.terminate_on_exception():
            if msg != WM_NCCREATE:
                if self_prop := user.get_prop_def(hwnd, _PROP_SELF):
                    self = ctypes.cast(self_prop, ctypes.py_object).value
                    res = self.on_message(msg, wp, lp)
                    if msg == WM_NCDESTROY:
                        self.del_prop(_PROP_SELF)
                        self.hwnd = None
                    return res

                # Some kind of messages may arrive before WM_NCCREATE
                # (e.g. WM_GETMINMAXINFO), i.e. still self_prop == None.
                return user.DefWindowProc(hwnd, msg, wp, lp)

            cparam = user.CREATESTRUCT.from_address(lp).lpCreateParams
            self = ctypes.cast(cparam, ctypes.py_object).value
            if isinstance(self, SimpleWnd):
                self.hwnd = hwnd
                self.set_prop(_PROP_SELF, cparam)
                return self.on_message(msg, wp, lp)
            raise TypeError("not derived from SimpleWnd")

    ############################################################################

    def create(self, wcp):
        if self.is_window():
            raise RecursionError("can only be created once")
        self.parent = wcp.parent
        wcp.cls.lpfnWndProc = self._wnd_proc_

        if wcp.cls.lpszClassName is None:
            # calc hash over wcp.cls and cast it to unsigned
            h = hash(bytes(wcp.cls)) & (2 ** sys.hash_info.width - 1)
            wcp.cls.lpszClassName = f"ctwin32:{h:x}"

        try:
            user.GetClassInfo(wcp.cls.hInstance, wcp.cls.lpszClassName)
        except OSError:
            user.RegisterClass(wcp.cls)

        user.CreateWindowEx(
            wcp.ex_style,
            wcp.cls.lpszClassName,
            wcp.name,
            wcp.wnd_style,
            wcp.left,
            wcp.top,
            wcp.width,
            wcp.height,
            wcp.parent,
            wcp.menu,
            wcp.cls.hInstance,
            id(self)
            )

    ############################################################################

    def on_message(self, msg, wp, lp):
        raise NotImplementedError("must be implemented in derived class")

    ############################################################################

    def __del__(self):
        if self.is_window():
            self.destroy()

################################################################################

def to_lzb85(raw_data):
    return base64.b85encode(lzma.compress(raw_data))

################################################################################

def from_lzb85(lzb85_data):
    return lzma.decompress(base64.b85decode(lzb85_data))

################################################################################

def load_ico_lz_b85(lz_b85_data):
    return user.CreateIconFromResourceEx(from_lzb85(lz_b85_data))

################################################################################

_py_icon = (
    b"{Wp48S^xk9=GL@E0stWa761SMbT8$j;1H(;>RkX705Aj)f+J6yP9T}Bko})j#+i7N>;e(Cc8"
    b"*b~t=p!4Q*Y-|8I`wx%g@F}t$1qY3`ER*6QY*}qNii+qAYq`NROBYh+Ot_RpNEGPRoAI1^Td"
    b"&(&<AnDap)j_o`+IK<rpFw3)<etl<IYLkGU|B2d7e`60A9qtCtFsg1{>=)qA*sCV&QUE12Nq"
    b"qR_6W0oGLTfny)d!CWv4njsF>CeO1THX!>_^ENrpO&Oy0;%fq$|!F3nno<dnv3(2y{c&itUa"
    b"1iK56!K;Z`9i*&0yl8zC$m9Tv+nv1#?r;S#KU<`4!FpMXj?Ht_jotwz*o?L-cp`Gc?6%uw_6"
    b"t-iWwadkQ7p<wFOI{#$S>sEa%fQbn&d5~&Tf->4?!WxGsv#=5n_b6v8kR;1Cxo}y}5chHkKd"
    b"+8<F2RVe^G^ZEH+XgIa*gbPH{u*_zX);IzvXAhgqBMc`{q6a%PeKwiU|3SKU_TZo;yWN+Jy5"
    b"{H{pQ+Q)RoAL)HBGYX-&-tgz0xCGDfYU0rZ3x-h+7onOS%J|hx^_kxhEy4c&JL<s`(xMxnMY"
    b"3?)MsDJGWRx;uosq;9nuFDfkJz-rEujXPpx0UITtH-mffmG<;y@p5W9+R*F!}6-#vKH-+lyX"
    b";J+#?QvOB7uPy4C53q`2Bt%8BHejw~~yQp1p{IKP_zJ3!Oq7|+)Ds@pyQ_64-$*ww>zN_2t5"
    b"o6k+215Gr)oYa5_P2;!g`oDgxp@9XFqjZ0x0m6TN5EZEvT?mtyP@QR*?VdVxiqkI%LqCL_04"
    b"8HcMK?Ej3afNs@Qs2L9gev&tW=$BmQbP*Gs+reQAWKocl|u*?5E^;`<qa9m|E&NKq1GAK$lC"
    b"4gp*7+h$GLJ;1nx8-No(Y%o*@?nBP8sE3D2`G3|+d)3Ul+wliQ~7A9gsR8N!>nL@g}ZSFIE<"
    b"3m(pl;5Ur1C9Mqcuyh-kf^e??xt{`a4E{mmds6mf6NUP&GKgkxYRM(g}FU5>lb>wTQ(m8dsL"
    b"{N-YuOa>4c3fm+n`!MpQ_(c6FmIq5-rI5_=(FhR{10ek%v;Pb!Y9)b1iKktnF@^w8;oP<S57"
    b"B-quZtTE=3L>_6TdPcc0A*;>4w-Y=gs%Z@-EDdv0g4m8a#V|5<Zt;?(@wYrVb-Mk1Zdw%Y09"
    b"Gn=e1&7;wA~N&ryFEU*V)J&^KvQY@Ak>Zm{0ooZsRK7?-#1(gu$iVqGdslP=0S|Q3<Fa-xec"
    b"8uu5r%nQ4UlsMx^Vni(7Bskl7aQdNJe>Db2}fWQ}$geIyG@D7n&OyNG7>c!_70Cy>lzy?L6p"
    b"<=%0-^CQkVI#_teeId#xO_L^xsUL?ycSIzlppqI74YA6(Zt6g{fFa6gQ(#f&Xs(UOr7@lz7L"
    b"Xc%_<xA_oa^vO}LkvmhHHq%Qe_z3^am;#oZ~$FtL3o@Cl{K^hF6+CUh+JE8-<96BRbQ2sw+0"
    b"07hXrCaq6ZRIn6iVjyQ^ufxm4gbLmXnMgIjh?7(|D&?Nlb2S%9rK@)Jz_irCdm`I{$-zZLXJ"
    b"v!RG$;h}KR@@r7eX*74~#4FYiyZBtxv8w`%IDomf2+mQ=p-v-~~BnB4ky$7MTefMN~%ZLb%`"
    b"9HAuZy>YC5wbuq}fsUv>XN>?<(K|l9W_i0*$<>pocq&1w+=2XX=ZTZ!b+9?|3)6=`R%F!}Z&"
    b"a3Qm#HB{fW|T^Vk!~MbAjB%uoKE2KB?<{2C4_7P^XpCPOvLvb*=`u!@U??9&!FwtogjvKchT"
    b"JYXIjeMr`}Q^G8vJ%n;<f;*!&FZ!Ihx~1V$NR`5Y3LJT5Gs{MWKOM$(1PgZT`{b##@>Y~@{s"
    b")vnA~VaZp}VEy{WN=0EzUX~N5K4PI_3hVd(<@S@?%#NnfV#o&CTj&LSlM;dAJ}n;7UKjmt%e"
    b"cG-MV+#m^)RYoItH4F4SpRv>o$*ci&39VZ{Gn)82#qhvYI77MY6k>v2iQs00000F``J)J91k"
    b"700D*!s38CVB-0#!vBYQl0ssI200dcD"
    )

def load_py_ico():
    return load_ico_lz_b85(_py_icon)

################################################################################

_ctwin32_icon = (
    b"{Wp48S^xk9=GL@E0stWa761SMbT8$j;0UJ#&0PQ#05Aj)f+J6yaHvHu1q>Gs(yZ)`rT4VqYn"
    b"%B-HIH~>jeV{SNI_{JHVRTYsiKzK7CwQ=tJ2Ks!~srhl-I*p0o!hI-fpo^97Xi*4Q!Dd3X}v"
    b"$8WKulmA{F!7oZCs)IAY;ZLMh6_No@b0ca=PkNzVlN3cC!ob60RrU0=4ndqroAy}qIF?$;ZH"
    b"Ufdr4QqNydBJoi**2(b>LJtsq1--PQVunbK*nFuD0^<0YG0-)S3tS_@5y)&cQ=s^k)-n~=oM"
    b"02WzYUu?pZecGjm&)ry+Orfv<jRMfx!kc1P+ahYuS(p%J>Y|E(NvE5ZQb(61)+uYPYcZKA?4"
    b"y3*?c3~|o@e>T)A&e2>Qsiwrg(x})20YiS)c@q~frD|ufg|rFed$7w!;pG&;vB0orztE80g1"
    b"5I$1~{%oA8kJqG8F5WFuF%de0KZlqk)<^C2K?eE3mj3mf16t&%ma@|F%17nc2|)O8_#Lldac"
    b"FKC#?m7Ga6mIU+oCQ4>>QHY^>Au-AQr)+1y^BmO-flIkHJsV)N);I;ebT<`hX9)46w<Rg3FL"
    b"i)nM%7pg~SO#qz-ze-1-*zMpUl#X<!?Q_$_PfG77v&?(kBe4H7IG?zCE?~%aW3h~O~Nk#)1H"
    b"v=0R1#>@CNEYYT5RzSq>D#w>P>#BfdIYc)Z?l^6ml~-ykX8X8NnIqKq+G;v$hN(Y<Jt)r?c9"
    b">GCHY0zJilG#Dgj6g&_Ar1vUBMuh7e1c=)&+mJs_S(AEH(w+z@wOPz`09ec$bD^<B20sPmtQ"
    b";9O8ejb}|4A@5yHoQ|la1UXI=fK{SBk8E0dNDC-7yw%?a9h+@fRDw!w6q>w{eR6p%f9?J2ey"
    b"sbWHE%E5LjOC=}=%xLInY)d36&`<eV73u+U8SBisPCILL$*_AbshRHS~A(Eo9HFi__M)NO{L"
    b"k}hiLF?7EhWwHZo$Xt984N&;?2^{2k^@GcRUbM$+nhqIJ0$;#L=XESY=_B_hRlCH6Hx-@ItE"
    b"@4C^*r-Z*y%%>?H81t0$s$1c97#vHl%HhMw%^^5Tid??!_#@o>WP7mN#K1i21pownz?8>#_R"
    b"67v0mi1&j!iRKMdj-fwjC}mJ4deM2@pDkT)?qtiS1a~80#D^R~9XM3@7QZL?zs-ZA-pCJo6{"
    b"R>GbdV}c`%kbaf1&ZpLT&u%w>OgdQZ*~7)q(fpFA=qdeAt|9=O5NX(*%bl8iYPJ(FJbra1Lc"
    b"1<m23X?h%S;!h}bzIodaWjLg}aLqJ_oc`_kb`^egLfTdLSI6#rFaK)MtiwsG%!qr%(+$M7=i"
    b"FsygaUfDcG1`f1)_y!(vw(U!A(ly-<MbTl&@rjG$BKX<x{({u@XJ*QsEUvq@f(bmc6v=U8!5"
    b"}p{K5@GE7Jm9Z!Bv3CFo#1ZHV7*nTNaA10bQO+6x22LetMJ2b#}!7h6#~GL*UD=`~w_>msdT"
    b"YdUl6{}zNFtK;1wkl9@$K|*;m;I}jOF`7rfqE9`8A2}5MAYGOm&Bd@13Bgk?!sLhcJ65PFvq"
    b"EK-oczoVl#{I->+Rapc9fQ>LT)>~&`%W3cz@D3$X)<L+qpZSQ4LslJyc~^rDkdX00000gHzk"
    b"KXFSSV00HR<s1X1FO9VA)vBYQl0ssI200dcD"
    )

def load_ctwin32_ico():
    return load_ico_lz_b85(_ctwin32_icon)

################################################################################

class BaseDlg(BaseWnd):

    def __init__(self, parent=None):
        super().__init__()
        self.parent = parent

    ############################################################################

    def parent_hwnd(self):
        return self.parent.hwnd if self.parent else None

    ############################################################################

    @user.DLGPROC
    @staticmethod
    def _dlg_proc_(hwnd, msg, wp, lp):
        # Since this is a python callback that ctypes calls when requested
        # by foreign C code, ctypes has no way of propagating any exception
        # back to the python interpreter - those would simply be ignored.
        # Therefore any exception has to terminate the process.
        with kernel.terminate_on_exception():
            if msg != WM_INITDIALOG:
                if self_prop := user.get_prop_def(hwnd, _PROP_SELF):
                    self = ctypes.cast(self_prop, ctypes.py_object).value
                    if msg == WM_COMMAND:
                        return self.on_command(
                            LOWORD(wp),
                            HIWORD(wp),
                            HWND(lp)
                            )
                    elif msg == WM_NOTIFY:
                        return self.on_notify(
                            UINT(wp).value,
                            ctypes.cast(lp, user.PNMHDR)
                            )
                    else:
                        if msg == WM_ACTIVATE and self.parent:
                            hdr = user.NMHDR(
                                hwnd,
                                user.GetDlgCtrlID(hwnd),
                                user.MSDN_ACTIVATE
                                )
                            ma = user.NM_MSD_ACTIVATE(hdr, wp != WA_INACTIVE)
                            self.parent.send_notify(ref(ma.hdr))
                        res = self.on_message(msg, wp, lp)
                        if (msg == WM_NCDESTROY):
                            self.del_prop(_PROP_SELF)
                            self.hwnd = None
                        return res
                else:
                    return False
            else:
                self = ctypes.cast(lp, ctypes.py_object).value
                if isinstance(self, BaseDlg):
                    self.hwnd = hwnd
                    self.set_prop(_PROP_SELF, lp)
                    return self.on_init_dialog()
                raise TypeError("not derived from BaseDlg")

    ############################################################################

    def create_modeless(self, template):
        return user.CreateDialogIndirectParam(
            template,
            self.parent,
            self._dlg_proc_,
            id(self)
            )

    ############################################################################

    def do_modal(self, template):
        return user.DialogBoxIndirectParam(
            template,
            self.parent,
            self._dlg_proc_,
            id(self)
            )

    ############################################################################

    def on_message(self, msg, wp, lp):
        return False

    def on_init_dialog(self):
        return True

    def on_command(self, cmd_id, notification, ctrl):
        user.EndDialog(self.hwnd, IDCANCEL)
        return True

    def on_notify(self, ctrl_id, pnmhdr):
        return False

    def __del__(self):
        self.destroy()

    def get_item(self, id):
        return self.get_dlg_item(id)

    def set_item_text(self, id, txt):
        return self.set_dlg_item_text(id, txt)

    def get_item_text(self, id):
        return self.get_dlg_item_text(id)

    def is_button_checked(self, id):
        return self.is_dlg_button_checked(id)

    def check_button(self, id, check):
        self.check_dlg_button(id, check)

    def send_destroy_request(self):
        if self.parent:
            md = user.NM_MSD_DESTROY(
                self.hwnd,
                user.GetDlgCtrlID(self.hwnd),
                user.MSDN_DESTROY
                )
            self.parent.send_notify(ref(md))
        else:
            self.destroy()

    def set_msg_result(self, result):
        user.SetWindowLongPtr(self.hwnd, DWLP_MSGRESULT, result)

################################################################################

_std_classes = {
    "button":    b"\xff\xff" + bytes(WORD(0x80)),
    "edit":      b"\xff\xff" + bytes(WORD(0x81)),
    "static":    b"\xff\xff" + bytes(WORD(0x82)),
    "listbox":   b"\xff\xff" + bytes(WORD(0x83)),
    "scrollbar": b"\xff\xff" + bytes(WORD(0x84)),
    "combobox":  b"\xff\xff" + bytes(WORD(0x85)),
    }

################################################################################

def dlg_item_template(
        style,
        x,
        y,
        cx,
        cy,
        id,
        cls,
        title,
        cdata=None
        ):
    style |= WS_VISIBLE | WS_CHILD
    # bytes = template + class + title + creation data
    cls = (
        _std_classes.get(cls.lower(), None) or
        bytes(string_buffer(cls))
        )
    cdata = (
        bytes(WORD(0)) if cdata is None
        else bytes(WORD(len(cdata))) + cdata
        )
    bts = (
        bytes(user.DLGITEMTEMPLATE(style, 0, x, y, cx, cy, id)) +
        cls +
        bytes(string_buffer(title)) +
        cdata
        )
    # align to DWORD for the following items
    bts += (-len(bts) % 4) * b"\0"
    assert len(bts) % 4 == 0
    return bts

################################################################################

def dlg_template(
        items,
        style,
        x,
        y,
        cx,
        cy,
        title,
        typeface,
        pointsize=8,
        ):
    # bytes = template + menu + class + title + pointsize + typeface + items
    style |= DS_SETFONT  # always set font
    tmpl = user.DLGTEMPLATE(style, 0, len(items), x, y, cx, cy)
    bts = b"".join((
        bytes(tmpl),
        bytes(WORD(0)),     # no menu
        bytes(WORD(0)),     # default class
        bytes(string_buffer(title)),
        bytes(WORD(pointsize)),
        bytes(string_buffer(typeface))
        ))
    return b"".join((
        bts,
        (-len(bts) % 4) * b"\0",  # align to DWORD for the following items
        *(dlg_item_template(*item) for item in items)
        ))

################################################################################

class InputDlg(BaseDlg):

    QUESTION_ID = 100
    ANSWER_ID   = 101
    ED_STYLE = ES_AUTOHSCROLL | ES_LEFT | WS_BORDER | WS_TABSTOP
    DLG_ITEMS = (
        (SS_CENTER | WS_GROUP, 7, 7, 201, 18, QUESTION_ID, "static", ""),
        (ED_STYLE, 7, 31, 201, 13, ANSWER_ID, "edit", ""),
        (BS_DEFPUSHBUTTON | WS_TABSTOP, 104, 51, 50, 14, IDOK, "button", "OK"),
        (WS_TABSTOP, 158, 51, 50, 14, IDCANCEL, "button", "Cancel"),
        )

    ############################################################################

    def __init__(self, fontsize=8, parent=None):
        self.fontsize = fontsize
        self.question = None
        self.answer = None
        self.password = False
        super().__init__(parent)

    ############################################################################

    def on_init_dialog(self):
        if self.password:
            # black circle -> U+25cf -> 9679
            self.get_item(self.ANSWER_ID).send_msg(EM_SETPASSWORDCHAR, 9679, 0)

        if self.question is not None:
            self.set_item_text(self.QUESTION_ID, self.question)

        self.center(self.parent)
        return True

    ############################################################################

    def on_command(self, cmd_id, notification, ctrl):
        if cmd_id in (IDOK, IDCANCEL):
            if cmd_id == IDOK:
                self.answer = self.get_item_text(self.ANSWER_ID)
            user.EndDialog(self.hwnd, cmd_id)
        return True

    ############################################################################

    def ask(self, question, caption="", password=False):
        self.question = question
        self.password = password
        template = dlg_template(
            self.DLG_ITEMS,
            DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU,
            0,
            0,
            215,
            72,
            caption,
            "MS Shell Dlg",
            (self.fontsize * self.get_dpi_scale_100() + 50) // 100
            )

        return self.answer if self.do_modal(template) == IDOK else None

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import sys
import time
import subprocess
from ctwin32 import kernel, UNICODE_STRING_MAX_CHARS

################################################################################

class NamedPipe:

    def __init__(self, name, *, server_end=False):
        self.name = name
        if server_end:
            self.server_end = True
            self.hdl = kernel.create_named_pipe(self.name)
        else:
            self.server_end = False
            self.hdl = kernel.create_file(self.name)

    def __enter__(self):
        return self

    def __exit__(self, typ, val, tb):
        self.hdl.close()

    def connect(self):
        if not self.server_end:
            raise ValueError("not server")
        kernel.ConnectNamedPipe(self.hdl)

    def disconnect(self):
        if not self.server_end:
            raise ValueError("not server")
        kernel.FlushFileBuffers(self.hdl)
        kernel.DisconnectNamedPipe(self.hdl)

    def read(self):
        return kernel.read_file_text(self.hdl, UNICODE_STRING_MAX_CHARS)

    def write(self, txt):
        kernel.write_file_text(self.hdl, txt)

################################################################################

PIPE_NAME = r"\\.\pipe\b52s_6060842"
CMD_STOP = "++stop++stop++stop++"
CMD_START = "++start++start++start++"

################################################################################

def run_server():

    print("server was started")

    with NamedPipe(PIPE_NAME, server_end=True) as pipe:
        while True:
            print("server is waiting for request")
            pipe.connect()
            print("server got client connection")

            try:
                txt = pipe.read()
            except BrokenPipeError:
                print("server broken pipe")
                pipe.disconnect()
                continue
            print(f"server received '{txt}'")
            pipe.write("".join(reversed(txt)))
            pipe.disconnect()
            if txt == CMD_STOP:
                print("server stops")
                break

################################################################################

def main(args):

    if args and args[0] == CMD_START:
        run_server()
        return

    try:
        print("client trying to open pipe")
        pipe = NamedPipe(PIPE_NAME)
    except FileNotFoundError:
        if not args:
            print("server not running, no args -> nothing to do")
            return
        print("starting server")
        subprocess.Popen([sys.executable, __file__, CMD_START])
        time.sleep(0.1)
        pipe = NamedPipe(PIPE_NAME)

    with pipe:
        if not args:
            args = [CMD_STOP]
        pipe.write(" ".join(args))
        txt = pipe.read()
        print(f"client received '{txt}'")

################################################################################

if __name__ == "__main__":
    main(sys.argv[1:])

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import ctypes
from .wtypes import (
    BOOL,
    DWORD,
    GUID,
    LONG,
    POINTER,
    PVOID,
    PWSTR,
    ULARGE_INTEGER,
    ULONG,
    )
from . import ref, fun_fact, raise_on_err

from .kernel import KHANDLE, PKHANDLE

_vdisk = ctypes.WinDLL("virtdisk.dll", use_last_error=True)

################################################################################

ATTACH_VIRTUAL_DISK_FLAG_NONE = 0
ATTACH_VIRTUAL_DISK_FLAG_READ_ONLY = 0x1
ATTACH_VIRTUAL_DISK_FLAG_NO_DRIVE_LETTER = 0x2
ATTACH_VIRTUAL_DISK_FLAG_PERMANENT_LIFETIME = 0x4
ATTACH_VIRTUAL_DISK_FLAG_NO_LOCAL_HOST = 0x8
ATTACH_VIRTUAL_DISK_FLAG_NO_SECURITY_DESCRIPTOR = 0x10
ATTACH_VIRTUAL_DISK_FLAG_BYPASS_DEFAULT_ENCRYPTION_POLICY = 0x20
ATTACH_VIRTUAL_DISK_FLAG_NON_PNP = 0x40
ATTACH_VIRTUAL_DISK_FLAG_RESTRICTED_RANGE = 0x80
ATTACH_VIRTUAL_DISK_FLAG_SINGLE_PARTITION = 0x100
ATTACH_VIRTUAL_DISK_FLAG_REGISTER_VOLUME = 0x200

VIRTUAL_DISK_ACCESS_NONE = 0
VIRTUAL_DISK_ACCESS_ATTACH_RO = 0x10000
VIRTUAL_DISK_ACCESS_ATTACH_RW = 0x20000
VIRTUAL_DISK_ACCESS_DETACH = 0x40000
VIRTUAL_DISK_ACCESS_GET_INFO = 0x80000
VIRTUAL_DISK_ACCESS_CREATE = 0x100000
VIRTUAL_DISK_ACCESS_METAOPS = 0x200000
VIRTUAL_DISK_ACCESS_READ = 0xd0000
VIRTUAL_DISK_ACCESS_ALL = 0x3f0000
VIRTUAL_DISK_ACCESS_WRITABLE = 0x320000

OPEN_VIRTUAL_DISK_FLAG_NONE = 0
OPEN_VIRTUAL_DISK_FLAG_NO_PARENTS = 0x1
OPEN_VIRTUAL_DISK_FLAG_BLANK_FILE = 0x2
OPEN_VIRTUAL_DISK_FLAG_BOOT_DRIVE = 0x4
OPEN_VIRTUAL_DISK_FLAG_CACHED_IO = 0x8
OPEN_VIRTUAL_DISK_FLAG_CUSTOM_DIFF_CHAIN = 0x10
OPEN_VIRTUAL_DISK_FLAG_PARENT_CACHED_IO = 0x20
OPEN_VIRTUAL_DISK_FLAG_VHDSET_FILE_ONLY = 0x40
OPEN_VIRTUAL_DISK_FLAG_IGNORE_RELATIVE_PARENT_LOCATOR = 0x80
OPEN_VIRTUAL_DISK_FLAG_NO_WRITE_HARDENING = 0x100
OPEN_VIRTUAL_DISK_FLAG_SUPPORT_COMPRESSED_VOLUMES = 0x200

OPEN_VIRTUAL_DISK_VERSION_UNSPECIFIED = 0
OPEN_VIRTUAL_DISK_VERSION_1 = 1
OPEN_VIRTUAL_DISK_VERSION_2 = 2
OPEN_VIRTUAL_DISK_VERSION_3 = 3

DETACH_VIRTUAL_DISK_FLAG_NONE = 0

ATTACH_VIRTUAL_DISK_VERSION_1 = 1
ATTACH_VIRTUAL_DISK_VERSION_2 = 2

################################################################################

class VIRTUAL_STORAGE_TYPE(ctypes.Structure):
    _fields_ = (
        ("DeviceId", ULONG),
        ("VendorId", GUID),
        )
PVIRTUAL_STORAGE_TYPE = POINTER(VIRTUAL_STORAGE_TYPE)

################################################################################

class _OVDP_VERSION1(ctypes.Structure):
    _fields_ = (
        ("RWDepth", ULONG),
        )

class _OVDP_VERSION2(ctypes.Structure):
    _fields_ = (
        ("GetInfoOnly", BOOL),
        ("ReadOnly", BOOL),
        ("ResiliencyGuid", GUID),
        )

class _OVDP_VERSION3(ctypes.Structure):
    _fields_ = (
        ("GetInfoOnly", BOOL),
        ("ReadOnly", BOOL),
        ("ResiliencyGuid", GUID),
        ("SnapshotId", GUID),
        )

class _OVDP_UNION(ctypes.Union):
    _fields_ = (
        ("Version1", _OVDP_VERSION1),
        ("Version2", _OVDP_VERSION2),
        ("Version3", _OVDP_VERSION3),
        )

class OPEN_VIRTUAL_DISK_PARAMETERS(ctypes.Structure):
    _fields_ = (("Version", LONG), ("u", _OVDP_UNION))
    _anonymous_ = ("u",)
POPEN_VIRTUAL_DISK_PARAMETERS = POINTER(OPEN_VIRTUAL_DISK_PARAMETERS)

################################################################################

class _AVDP_VERSION1(ctypes.Structure):
    _fields_ = (
        ("Reserved", ULONG),
        )

class _AVDP_VERSION2(ctypes.Structure):
    _fields_ = (
        ("RestrictedOffset", ULARGE_INTEGER),
        ("RestrictedLength", ULARGE_INTEGER),
        )

class _AVDP_UNION(ctypes.Union):
    _fields_ = (
        ("Version1", _OVDP_VERSION1),
        ("Version2", _OVDP_VERSION2),
        )

class ATTACH_VIRTUAL_DISK_PARAMETERS(ctypes.Structure):
    _fields_ = (("Version", LONG), ("u", _AVDP_UNION))
    _anonymous_ = ("u",)
PATTACH_VIRTUAL_DISK_PARAMETERS = POINTER(ATTACH_VIRTUAL_DISK_PARAMETERS)

################################################################################

_OpenVirtualDisk = fun_fact(
    _vdisk.OpenVirtualDisk, (
        DWORD,
        PVIRTUAL_STORAGE_TYPE,
        PWSTR,
        LONG,
        LONG,
        POPEN_VIRTUAL_DISK_PARAMETERS,
        PKHANDLE
        )
    )

def OpenVirtualDisk(storage_type, path, access_mask, flags, parameters=None):
    hdl = KHANDLE()
    raise_on_err(
        _OpenVirtualDisk(
            ref(storage_type),
            path,
            access_mask,
            flags,
            None if parameters is None else ref(parameters),
            ref(hdl)
            )
        )
    return hdl

################################################################################

_AttachVirtualDisk = fun_fact(
    _vdisk.AttachVirtualDisk, (
        DWORD,
        KHANDLE,
        PVOID,  # no interest in supplying a security descriptor
        LONG,
        ULONG,
        PATTACH_VIRTUAL_DISK_PARAMETERS,
        PVOID,  # no interest in supplying an overlapped
        )
    )

def AttachVirtualDisk(hdl, flags, prov_flags=0, parameters=None):
    raise_on_err(
        _AttachVirtualDisk(
            hdl,
            None,
            flags,
            prov_flags,
            None if parameters is None else ref(parameters),
            None
            )
        )

################################################################################

_DetachVirtualDisk = fun_fact(
    _vdisk.DetachVirtualDisk, (
        DWORD,
        KHANDLE,
        LONG,
        ULONG,
        )
    )

def DetachVirtualDisk(hdl, flags=0, prov_flags=0):
    raise_on_err(_DetachVirtualDisk(hdl, flags, prov_flags))

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

from .wtypes import (
    BOOLEAN,
    BYTE,
    ctypes,
    INT,
    LARGE_INTEGER,
    LONG,
    LONGLONG,
    ULONG,
    ULONGLONG,
    USHORT,
    WCHAR,
    )

################################################################################

class KSYSTEM_TIME(ctypes.Structure):
    _fields_ = (
        ("LowPart", ULONG),
        ("High1Time", LONG),
        ("High2Time", LONG),
        )

class XSTATE_FEATURE(ctypes.Structure):
    _fields_ = (
        ("Offset", ULONG),
        ("Size", ULONG),
        )

MAXIMUM_XSTATE_FEATURES = 64
PROCESSOR_FEATURE_MAX = 64

class XSTATE_CONFIGURATION(ctypes.Structure):
    _fields_ = (
        ("EnabledFeatures", ULONGLONG),
        ("EnabledVolatileFeatures", ULONGLONG),
        ("Size", ULONG),
        ("ControlFlags", ULONG),
        ("Features", XSTATE_FEATURE * MAXIMUM_XSTATE_FEATURES),
        ("EnabledSupervisorFeatures", ULONGLONG),
        ("AlignedFeatures", ULONGLONG),
        ("AllFeatureSize", ULONG),
        ("AllFeatures", ULONG * MAXIMUM_XSTATE_FEATURES),
        ("EnabledUserVisibleSupervisorFeatures", ULONGLONG),
        ("ExtendedFeatureDisableFeatures", ULONGLONG),
        ("AllNonLargeFeatureSize", ULONG),
        ("Spare", ULONG),
        )

################################################################################

class KUSER_SHARED_DATA(ctypes.Structure):
    _fields_ = (
        ("TickCountLowDeprecated", ULONG),
        ("TickCountMultiplier", ULONG),
        ("InterruptTime", KSYSTEM_TIME),
        ("SystemTime", KSYSTEM_TIME),
        ("TimeZoneBias", KSYSTEM_TIME),
        ("ImageNumberLow", USHORT),
        ("ImageNumberHigh", USHORT),
        ("NtSystemRoot", WCHAR * 260),
        ("MaxStackTraceDepth", ULONG),
        ("CryptoExponent", ULONG),
        ("TimeZoneId", ULONG),
        ("LargePageMinimum", ULONG),
        ("AitSamplingValue", ULONG),
        ("AppCompatFlag", ULONG),
        ("RNGSeedVersion", ULONGLONG),
        ("GlobalValidationRunlevel", ULONG),
        ("TimeZoneBiasStamp", LONG),
        ("NtBuildNumber", ULONG),
        ("NtProductType", INT),
        ("ProductTypeIsValid", BOOLEAN),
        ("Reserved0", BOOLEAN),
        ("NativeProcessorArchitecture", USHORT),
        ("NtMajorVersion", ULONG),
        ("NtMinorVersion", ULONG),
        ("ProcessorFeatures", BOOLEAN * PROCESSOR_FEATURE_MAX),
        ("Reserved1", ULONG),
        ("Reserved3", ULONG),
        ("TimeSlip", ULONG),
        ("AlternativeArchitecture", INT),
        ("BootId", ULONG),
        ("SystemExpirationDate", LARGE_INTEGER),
        ("SuiteMask", ULONG),
        ("KdDebuggerEnabled", BOOLEAN),
        ("MitigationPolicies", BYTE),
        ("CyclesPerYield", USHORT),
        ("ActiveConsoleId", ULONG),
        ("DismountCount", ULONG),
        ("ComPlusPackage", ULONG),
        ("LastSystemRITEventTickCount", ULONG),
        ("NumberOfPhysicalPages", ULONG),
        ("SafeBootMode", BOOLEAN),
        ("VirtualizationFlags", BYTE),
        ("Reserved12", BYTE * 2),
        ("SharedDataFlags", ULONG),
        ("DataFlagsPad", ULONG),
        ("TestRetInstruction", ULONGLONG),
        ("QpcFrequency", LONGLONG),
        ("SystemCall", ULONG),
        ("Reserved2", ULONG),
        ("SystemCallPad", ULONGLONG * 2),
        ("TickCountQuad", ULONGLONG),
        ("TickCountPad", ULONGLONG),
        ("Cookie", ULONG),
        ("CookiePad", ULONG),
        ("ConsoleSessionForegroundProcessId", LONGLONG),
        ("TimeUpdateLock", ULONGLONG),
        ("BaselineSystemTimeQpc", ULONGLONG),
        ("BaselineInterruptTimeQpc", ULONGLONG),
        ("QpcSystemTimeIncrement", ULONGLONG),
        ("QpcInterruptTimeIncrement", ULONGLONG),
        ("QpcSystemTimeIncrementShift", BYTE),
        ("QpcInterruptTimeIncrementShift", BYTE),
        ("UnparkedProcessorCount", USHORT),
        ("EnclaveFeatureMask", ULONG * 4),
        ("TelemetryCoverageRound", ULONG),
        ("UserModeGlobalLogger", USHORT * 16),
        ("ImageFileExecutionOptions", ULONG),
        ("LangGenerationCount", ULONG),
        ("Reserved4", ULONGLONG),
        ("InterruptTimeBias", ULONGLONG),
        ("QpcBias", ULONGLONG),
        ("ActiveProcessorCount", ULONG),
        ("ActiveGroupCount", BYTE),
        ("Reserved9", BYTE),
        ("QpcData", USHORT),
        ("TimeZoneBiasEffectiveStart", LARGE_INTEGER),
        ("TimeZoneBiasEffectiveEnd", LARGE_INTEGER),
        ("XState", XSTATE_CONFIGURATION),
        ("FeatureConfigurationChangeStamp", KSYSTEM_TIME),
        ("Spare", ULONG),
        ("UserPointerAuthMask", ULONGLONG),
    )

KUSER_SHARED_DATA_ADDRESS = 0x7FFE0000

################################################################################

def get_ref():
    return KUSER_SHARED_DATA.from_address(KUSER_SHARED_DATA_ADDRESS)

################################################################################

def get_data():
    # have to make a copy in order to actually adopt the data NOW!
    return KUSER_SHARED_DATA.from_buffer_copy(get_ref())

################################################################################

def _validate():
    assert KUSER_SHARED_DATA.TickCountLowDeprecated.offset == 0x0
    assert KUSER_SHARED_DATA.TickCountMultiplier.offset == 0x4
    assert KUSER_SHARED_DATA.InterruptTime.offset == 0x08
    assert KUSER_SHARED_DATA.SystemTime.offset == 0x014
    assert KUSER_SHARED_DATA.TimeZoneBias.offset == 0x020
    assert KUSER_SHARED_DATA.ImageNumberLow.offset == 0x02c
    assert KUSER_SHARED_DATA.ImageNumberHigh.offset == 0x02e
    assert KUSER_SHARED_DATA.NtSystemRoot.offset == 0x030
    assert KUSER_SHARED_DATA.MaxStackTraceDepth.offset == 0x238
    assert KUSER_SHARED_DATA.CryptoExponent.offset == 0x23c
    assert KUSER_SHARED_DATA.TimeZoneId.offset == 0x240
    assert KUSER_SHARED_DATA.LargePageMinimum.offset == 0x244
    assert KUSER_SHARED_DATA.AitSamplingValue.offset == 0x248
    assert KUSER_SHARED_DATA.AppCompatFlag.offset == 0x24c
    assert KUSER_SHARED_DATA.RNGSeedVersion.offset == 0x250
    assert KUSER_SHARED_DATA.GlobalValidationRunlevel.offset == 0x258
    assert KUSER_SHARED_DATA.TimeZoneBiasStamp.offset == 0x25c
    assert KUSER_SHARED_DATA.NtBuildNumber.offset == 0x260
    assert KUSER_SHARED_DATA.NtProductType.offset == 0x264
    assert KUSER_SHARED_DATA.ProductTypeIsValid.offset == 0x268
    assert KUSER_SHARED_DATA.NativeProcessorArchitecture.offset == 0x26a
    assert KUSER_SHARED_DATA.NtMajorVersion.offset == 0x26c
    assert KUSER_SHARED_DATA.NtMinorVersion.offset == 0x270
    assert KUSER_SHARED_DATA.ProcessorFeatures.offset == 0x274
    assert KUSER_SHARED_DATA.Reserved1.offset == 0x2b4
    assert KUSER_SHARED_DATA.Reserved3.offset == 0x2b8
    assert KUSER_SHARED_DATA.TimeSlip.offset == 0x2bc
    assert KUSER_SHARED_DATA.AlternativeArchitecture.offset == 0x2c0
    assert KUSER_SHARED_DATA.SystemExpirationDate.offset == 0x2c8
    assert KUSER_SHARED_DATA.SuiteMask.offset == 0x2d0
    assert KUSER_SHARED_DATA.KdDebuggerEnabled.offset == 0x2d4
    assert KUSER_SHARED_DATA.MitigationPolicies.offset == 0x2d5
    assert KUSER_SHARED_DATA.CyclesPerYield.offset == 0x2d6
    assert KUSER_SHARED_DATA.ActiveConsoleId.offset == 0x2d8
    assert KUSER_SHARED_DATA.DismountCount.offset == 0x2dc
    assert KUSER_SHARED_DATA.ComPlusPackage.offset == 0x2e0
    assert KUSER_SHARED_DATA.LastSystemRITEventTickCount.offset == 0x2e4
    assert KUSER_SHARED_DATA.NumberOfPhysicalPages.offset == 0x2e8
    assert KUSER_SHARED_DATA.SafeBootMode.offset == 0x2ec
    assert KUSER_SHARED_DATA.VirtualizationFlags.offset == 0x2ed
    assert KUSER_SHARED_DATA.Reserved12.offset == 0x2ee
    assert KUSER_SHARED_DATA.SharedDataFlags.offset == 0x2f0
    assert KUSER_SHARED_DATA.TestRetInstruction.offset == 0x2f8
    assert KUSER_SHARED_DATA.QpcFrequency.offset == 0x300
    assert KUSER_SHARED_DATA.SystemCall.offset == 0x308
    assert KUSER_SHARED_DATA.Reserved2.offset == 0x30c
    assert KUSER_SHARED_DATA.SystemCallPad.offset == 0x310
    assert KUSER_SHARED_DATA.TickCountQuad.offset == 0x320
    assert KUSER_SHARED_DATA.Cookie.offset == 0x330
    assert KUSER_SHARED_DATA.ConsoleSessionForegroundProcessId.offset == 0x338
    assert KUSER_SHARED_DATA.TimeUpdateLock.offset == 0x340
    assert KUSER_SHARED_DATA.BaselineSystemTimeQpc.offset == 0x348
    assert KUSER_SHARED_DATA.BaselineInterruptTimeQpc.offset == 0x350
    assert KUSER_SHARED_DATA.QpcSystemTimeIncrement.offset == 0x358
    assert KUSER_SHARED_DATA.QpcInterruptTimeIncrement.offset == 0x360
    assert KUSER_SHARED_DATA.QpcSystemTimeIncrementShift.offset == 0x368
    assert KUSER_SHARED_DATA.QpcInterruptTimeIncrementShift.offset == 0x369
    assert KUSER_SHARED_DATA.UnparkedProcessorCount.offset == 0x36a
    assert KUSER_SHARED_DATA.EnclaveFeatureMask.offset == 0x36c
    assert KUSER_SHARED_DATA.TelemetryCoverageRound.offset == 0x37c
    assert KUSER_SHARED_DATA.UserModeGlobalLogger.offset == 0x380
    assert KUSER_SHARED_DATA.ImageFileExecutionOptions.offset == 0x3a0
    assert KUSER_SHARED_DATA.LangGenerationCount.offset == 0x3a4
    assert KUSER_SHARED_DATA.Reserved4.offset == 0x3a8
    assert KUSER_SHARED_DATA.InterruptTimeBias.offset == 0x3b0
    assert KUSER_SHARED_DATA.QpcBias.offset == 0x3b8
    assert KUSER_SHARED_DATA.ActiveProcessorCount.offset == 0x3c0
    assert KUSER_SHARED_DATA.ActiveGroupCount.offset == 0x3c4
    assert KUSER_SHARED_DATA.Reserved9.offset == 0x3c5
    assert KUSER_SHARED_DATA.QpcData.offset == 0x3c6
    assert KUSER_SHARED_DATA.TimeZoneBiasEffectiveStart.offset == 0x3c8
    assert KUSER_SHARED_DATA.TimeZoneBiasEffectiveEnd.offset == 0x3d0
    assert KUSER_SHARED_DATA.XState.offset == 0x3d8
    assert KUSER_SHARED_DATA.FeatureConfigurationChangeStamp.offset == 0x720
    assert KUSER_SHARED_DATA.UserPointerAuthMask.offset == 0x730
    assert ctypes.sizeof(KUSER_SHARED_DATA) == 0x738

################################################################################

def _demo():
    _validate()
    from pprint import pprint # noqa: PLC0415
    from ctwin32 import ns_from_struct # noqa: PLC0415

    pprint(ns_from_struct(get_ref()))

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import ctypes
from .wtypes import (
    string_buffer,
    BOOL,
    DWORD,
    HANDLE,
    HMODULE,
    PDWORD,
    POINTER,
    PVOID,
    PWSTR,
    WinError,
    )
from . import (
    ref,
    ntdll,
    kernel,
    raise_on_zero,
    fun_fact,
    ns_from_struct,
    LIST_MODULES_DEFAULT,
    ERROR_PARTIAL_COPY,
    )

_psa = ctypes.WinDLL("psapi.dll", use_last_error=True)

################################################################################

_EnumProcessModulesEx = fun_fact(
    _psa.EnumProcessModulesEx, (BOOL, HANDLE, PVOID, DWORD, PDWORD, DWORD)
    )

def EnumProcessModulesEx(hdl, filter):
    HSIZE = ctypes.sizeof(HMODULE)
    needed = DWORD(128 * HSIZE)
    size = DWORD(0)
    while needed.value > size.value:
        size.value = needed.value
        mods = (HMODULE * (size.value // HSIZE))()
        if not _EnumProcessModulesEx(hdl, ref(mods), size, ref(needed), filter):
            if (err := kernel.GetLastError()) != ERROR_PARTIAL_COPY:
                raise WinError(err)
    return mods[:needed.value // HSIZE]

################################################################################

def EnumProcessModules(hdl):
    return EnumProcessModulesEx(hdl, LIST_MODULES_DEFAULT)

################################################################################

def EnumProcesses():
    return [p.pid for p in ntdll.enum_processes()]

################################################################################

_GetMappedFileName = fun_fact(
    _psa.GetMappedFileNameW, (DWORD, HANDLE, PVOID, PWSTR, DWORD)
    )

def GetMappedFileName(hdl, addr):
    size = 128
    length = size
    while length == size:
        size *= 2
        name = string_buffer(size)
        length = _GetMappedFileName(hdl, addr, name, size)
    return ntdll._resolve_device_prefix(name.value)

################################################################################

_GetModuleFileNameEx = fun_fact(
    _psa.GetModuleFileNameExW, (DWORD, HANDLE, PVOID, PWSTR, DWORD)
    )

def GetModuleFileNameEx(hdl, mod):
    size = 128
    length = size
    while length == size:
        size *= 2
        name = string_buffer(size)
        length = _GetModuleFileNameEx(hdl, mod, name, size)
        raise_on_zero(length)
    return name.value

################################################################################

class MODULEINFO(ctypes.Structure):
    _fields_ = (
        ("lpBaseOfDll", PVOID),
        ("SizeOfImage", DWORD),
        ("EntryPoint", PVOID),
        )

PMODULEINFO = POINTER(MODULEINFO)

_GetModuleInformation = fun_fact(
    _psa.GetModuleInformation, (BOOL, HANDLE, PVOID, PMODULEINFO, DWORD)
    )

def GetModuleInformation(hdl, mod):
    info = MODULEINFO()
    raise_on_zero(
        _GetModuleInformation(hdl, mod, ref(info), ctypes.sizeof(info))
        )
    return ns_from_struct(info)

################################################################################

def GetProcessImageFileName(hdl):
    return ntdll.proc_path_from_handle(hdl)

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################
#
# This sample illustrates the NT object namespace by printing the contents of
# of the root directory recursively,
#
################################################################################

from ctwin32 import (
    ntdll,
    suppress_winerr,
    DIRECTORY_QUERY,
    ERROR_ACCESS_DENIED,
    GENERIC_READ,
    )

################################################################################

def open_obj_dir(dir_name, access):
    return ntdll.NtOpenDirectoryObject(access, ntdll.obj_attr(dir_name))

################################################################################

def directory_contents(dir_name):
    with open_obj_dir(dir_name, DIRECTORY_QUERY) as hdir:
        return ntdll.NtQueryDirectoryObject(hdir)

################################################################################

def resolve_link(dir_name, link_name):
    with suppress_winerr(ERROR_ACCESS_DENIED):
        with open_obj_dir(dir_name, GENERIC_READ) as hdir:
            al = ntdll.obj_attr(link_name, hdir)
            with ntdll.NtOpenSymbolicLinkObject(GENERIC_READ, al) as hl:
                return ntdll.NtQuerySymbolicLinkObject(hl)
    return ""

################################################################################

def print_directory(dname, indent=""):
    try:
        info = directory_contents(dname)
    except OSError as e:
        if e.winerror == ERROR_ACCESS_DENIED:
            print(f"{indent}access denied (try running elevated)")
            return
        raise

    for oname, typ in info:
        lnk = "" if typ != "SymbolicLink" else resolve_link(dname, oname)
        print(f"{indent}{oname:<30} {typ:<20} {lnk}")
        if typ == "Directory":
            sub = f"{dname}\\{oname}" if len(dname) > 1 else f"\\{oname}"
            print_directory(sub, indent + "    ")

################################################################################

if __name__ == "__main__":
    print_directory("\\")

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################
# This sample was inspired by
# https://www.codeproject.com/Articles/13839/How-to-Prepare-a-USB-Drive-for-Safe-Removal
#
# It demonstrates how to prepare a disk drive for safe removal. The easy part
# is calling CM_Request_Device_Eject. The difficult one is to find the devinst
# that corresponds to a certain drive letter. There are lots cases where this
# will NOT work (e.g. mounted ISO, VHD or VeraCrypt volumes). In fact, it is
# a oversimplification to assume that there is always a one-to-one relationship
# between drive letter and disk device. But for the simple cases like USB
# sticks or USB harddisks this should work.
#
################################################################################

import sys
import time
import ctypes

from ctwin32 import (
    kernel,
    cfgmgr,
    setupapi,
    FILE_DEVICE_CD_ROM,
    FILE_DEVICE_DVD,
    FILE_DEVICE_DISK,
    )
from ctwin32.wtypes import INT, ULONG, GUID

################################################################################

# from winioctl.h

class STORAGE_DEVICE_NUMBER(ctypes.Structure):
    _fields_ = (
        ("DeviceType", INT),
        ("DeviceNumber", ULONG),
        ("PartitionNumber", ULONG),
        )

IOCTL_STORAGE_GET_DEVICE_NUMBER = 0x002d1080
GUID_IFACE_DISK = GUID("53f56307-b6bf-11d0-94f2-00a0c91efb8b")
GUID_IFACE_CDROM = GUID("53f56308-b6bf-11d0-94f2-00a0c91efb8b")

################################################################################

def get_drive_type_number(file_name):
    with kernel.create_file(file_name, 0) as vol:
        sdn = kernel.DeviceIoControl(
            vol,
            IOCTL_STORAGE_GET_DEVICE_NUMBER,
            None,
            ctypes.sizeof(STORAGE_DEVICE_NUMBER)
            )
    sdn = STORAGE_DEVICE_NUMBER.from_buffer(sdn)
    return sdn.DeviceType, sdn.DeviceNumber

################################################################################

def get_drive_devinst(drv_type, drv_num):
    # In order to find the devinst that corresponds to a certain drive, we
    # have to iterate over all device interfaces that match the drive type. The
    # drive number is only unique within the group of devices that share the
    # same interface.
    if drv_type in (FILE_DEVICE_CD_ROM, FILE_DEVICE_DVD):
        guid = GUID_IFACE_CDROM
    elif drv_type == FILE_DEVICE_DISK:
        # ignoring that floppies ever existed
        guid = GUID_IFACE_DISK
    else:
        raise ValueError(f"unhandled drive type: {drv_type}")

    for iset, did in setupapi.enum_dev_interfaces(guid):
        dev_path, deinda = setupapi.SetupDiGetDeviceInterfaceDetail(iset, did)
        dtype, dnum = get_drive_type_number(dev_path)
        if dtype == drv_type and dnum == drv_num:
            return deinda.DevInst

    raise OSError(f"devinst not found for {drv_type} {drv_num}")

################################################################################

def remove_drive_by_letter(letter):
    o = ord(letter)
    if o < ord("A") or o > ord("Z"):
        raise ValueError(f"invalid drive letter: {letter}")

    drv_type, drv_num = get_drive_type_number(f"\\\\.\\{letter}:")
    devinst = get_drive_devinst(drv_type, drv_num)
    parent = cfgmgr.CM_Get_Parent(devinst)

    MAX_TRIES = 3
    for i in range(MAX_TRIES):
        try:
            cfgmgr.CM_Request_Device_Eject(parent)
            err = None
            break
        except OSError as e:
            err = e
        if i < MAX_TRIES - 1:
            time.sleep(0.5)

    if err is None:
        print("--> Sucess <--")
    else:
        print(err)

################################################################################

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("need drive letter")
    else:
        remove_drive_by_letter(sys.argv[1][0].upper())

################################################################################
